# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from infoblox_client import exceptions as ibc_exc
from infoblox_client import objects as obj
from neutron_lib import constants as n_const
from oslo_log import log as logging
from oslo_utils import excutils

from networking_infoblox.neutron.common import constants
from networking_infoblox.neutron.common import ea_manager as eam
from networking_infoblox.neutron.common import pattern
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

LOG = logging.getLogger(__name__)


class DnsController(object):

    def __init__(self, ib_context):
        self.ib_cxt = ib_context
        self.grid_config = self.ib_cxt.grid_config
        self.grid_id = self.grid_config.grid_id
        self.pattern_builder = pattern.PatternBuilder(self.ib_cxt)
        self.dns_zone = self.pattern_builder.get_zone_name(
            is_external=self.ib_cxt.network_is_external)
        self._update_strategy_and_eas()

    def _update_strategy_and_eas(self):
        self.need_forward = (constants.ZONE_CREATION_STRATEGY_FORWARD in
                             self.grid_config.zone_creation_strategy)
        if self.need_forward:
            self.forward_zone_eas = eam.get_ea_for_forward_zone(
                self.ib_cxt.user_id, self.ib_cxt.tenant_id,
                self.ib_cxt.tenant_name, self.ib_cxt.network,
                self.ib_cxt.subnet,
                self.pattern_builder.get_zone_name(
                    is_external=self.ib_cxt.network_is_external))
        self.need_reverse = (constants.ZONE_CREATION_STRATEGY_REVERSE in
                             self.grid_config.zone_creation_strategy)
        if self.need_reverse:
            self.reverse_zone_eas = eam.get_ea_for_reverse_zone(
                self.ib_cxt.user_id, self.ib_cxt.tenant_id,
                self.ib_cxt.tenant_name, self.ib_cxt.network,
                self.ib_cxt.subnet)

    def create_dns_zones(self, rollback_list):
        if self.grid_config.dns_support is False:
            return
        LOG.debug("Initialized DNS Zone creation")
        cidr = self.ib_cxt.subnet['cidr']
        subnet_name = self.ib_cxt.subnet['name']
        dns_view = self.ib_cxt.mapping.dns_view
        ns_group = self.grid_config.ns_group
        zone_format = "IPV%s" % self.ib_cxt.subnet['ip_version']
        prefix = utils.get_ipv4_network_prefix(cidr, subnet_name)

        grid_primaries, grid_secondaries = self.ib_cxt.get_dns_members()

        if ns_group:
            # create Forward zone
            if self.need_forward:
                ib_zone, obj_created = obj.DNSZone.create_check_exists(
                    self.ib_cxt.connector,
                    view=dns_view,
                    ns_group=ns_group,
                    fqdn=self.dns_zone,
                    extattrs=self.forward_zone_eas)
                if ib_zone and obj_created:
                    LOG.info("Created forward zone: %s with ns_group: %s" % (
                             self.dns_zone, ns_group))
                    rollback_list.append(ib_zone)

            # create Reverse zone
            if self.need_reverse:
                ib_zone_cidr, obj_created = obj.DNSZone.create_check_exists(
                    self.ib_cxt.connector,
                    view=dns_view,
                    fqdn=cidr,
                    prefix=prefix,
                    zone_format=zone_format,
                    extattrs=self.reverse_zone_eas)
                if ib_zone_cidr and obj_created:
                    LOG.info("Created reverse zone: %s with ns_group: %s" % (
                             cidr, ns_group))
                    rollback_list.append(ib_zone_cidr)
        else:
            # create Forward zone
            if self.need_forward:
                ib_zone, obj_created = obj.DNSZone.create_check_exists(
                    self.ib_cxt.connector,
                    view=dns_view,
                    fqdn=self.dns_zone,
                    grid_primary=grid_primaries,
                    grid_secondaries=grid_secondaries,
                    extattrs=self.forward_zone_eas)
                if ib_zone and obj_created:
                    LOG.info("Created forward zone: %s with "
                             "grid_primaries: %s, grid_secondaries: %s" % (
                                 self.dns_zone, grid_primaries,
                                 grid_secondaries))
                    rollback_list.append(ib_zone)

            # create Reverse zone
            if self.need_reverse:
                ib_zone_cidr, obj_created = obj.DNSZone.create_check_exists(
                    self.ib_cxt.connector,
                    view=dns_view,
                    fqdn=cidr,
                    grid_primary=grid_primaries,
                    prefix=prefix,
                    zone_format=zone_format,
                    extattrs=self.reverse_zone_eas)
                if ib_zone_cidr and obj_created:
                    LOG.info("Created reverse zone: %s with "
                             "grid_primaries: %s, grid_secondaries: %s" % (
                                 cidr, grid_primaries, grid_secondaries))
                    rollback_list.append(ib_zone_cidr)

    def update_dns_zones(self):
        if self.grid_config.dns_support is False:
            return

        dns_view = self.ib_cxt.mapping.dns_view

        # update Forward zone
        if self.need_forward:
            self.ib_cxt.ibom.update_dns_zone_attrs(
                dns_view, self.dns_zone, self.forward_zone_eas)
        # update Reverse zone
        if self.need_reverse:
            self.ib_cxt.ibom.update_dns_zone_attrs(
                dns_view, self.ib_cxt.subnet['cidr'], self.reverse_zone_eas)

    def delete_dns_zones(self, dns_zone=None, ib_network=None):
        if self.grid_config.dns_support is False:
            return

        session = self.ib_cxt.context.session
        dns_view = self.ib_cxt.mapping.dns_view
        cidr = self.ib_cxt.subnet['cidr']
        dns_zone = dns_zone if dns_zone else self.dns_zone

        db_network_views = dbi.get_network_views(
            session,
            network_view_id=self.ib_cxt.mapping.network_view_id,
            participated=True)
        if not db_network_views:
            LOG.info("Network view has been removed so dns zone is already"
                     "removed.")
            return

        zone_removable = (not self.ib_cxt.network_is_shared_or_external or
                          self.grid_config.admin_network_deletion)
        if zone_removable:
            # delete forward zone
            if self._is_forward_zone_removable():
                self.ib_cxt.ibom.delete_dns_zone(dns_view, dns_zone)

            # delete reverse zone
            self.ib_cxt.ibom.delete_dns_zone(dns_view, cidr)

        # for external/shared network
        # the zone could be fixed at "cloud.infoblox.com" and so just deleting
        # a network that used that zone causing the EAs to clear on the zone
        # wouldn't be right.

    def _is_forward_zone_removable(self):
        session = self.ib_cxt.context.session
        pattern = self.grid_config.default_domain_name_pattern

        subnet_id = self.ib_cxt.subnet['id']
        network_id = self.ib_cxt.subnet['network_id']
        tenant_id = self.ib_cxt.subnet['tenant_id']

        # check all dynamic patterns from bottom to top hierarchy
        subnet_used = '{subnet_name}' in pattern or '{subnet_id}' in pattern
        if subnet_used:
            return True

        network_used = '{network_name}' in pattern or '{network_id}' in pattern
        if network_used:
            if dbi.is_last_subnet_in_network(session, subnet_id, network_id):
                return True

        tenant_used = '{tenant_name}' in pattern or '{tenant_id}' in pattern
        if tenant_used:
            if dbi.is_last_subnet_in_tenant(session, subnet_id, tenant_id):
                return True

        address_scope_used = ('{address_scope_name}' in pattern or
                              '{address_scope_id}' in pattern)
        if address_scope_used:
            if dbi.is_last_subnet_in_address_scope(session, subnet_id):
                return True

        # now check for static zone
        if self.grid_config.allow_static_zone_deletion:
            return dbi.is_last_subnet_in_private_networks(session, subnet_id)
        else:
            return False

    def bind_names(self, ip_address, instance_name=None, port_id=None,
                   port_tenant_id=None, device_id=None, device_owner=None,
                   is_floating_ip=False, port_name=None):
        if not device_owner:
            return

        if device_owner in constants.NEUTRON_DEVICE_OWNER_COMPUTE_LIST:
            if instance_name is None:
                instance = dbi.get_instance(self.ib_cxt.context.session,
                                            device_id)
                if instance is not None:
                    instance_name = instance.instance_name

        tenant_id = port_tenant_id or self.ib_cxt.context.tenant_id
        tenant_name = self.ib_cxt.get_tenant_name(tenant_id)
        ea_ip_address = eam.get_ea_for_ip(self.ib_cxt.user_id,
                                          tenant_id,
                                          tenant_name,
                                          self.ib_cxt.network,
                                          port_id,
                                          device_id,
                                          device_owner,
                                          is_floating_ip,
                                          instance_name)

        ip_alloc = (self.ib_cxt.dhcp_port_ip_alloc
                    if device_owner == n_const.DEVICE_OWNER_DHCP
                    else self.ib_cxt.ip_alloc)
        try:
            self._bind_names(ip_alloc.bind_names, ip_address,
                             instance_name, port_id,
                             port_tenant_id, device_id,
                             device_owner, ea_ip_address,
                             port_name, tenant_name)
        except ibc_exc.InfobloxCannotCreateObject:
            with excutils.save_and_reraise_exception():
                self.unbind_names(ip_address, instance_name, port_id,
                                  port_tenant_id, device_id, device_owner,
                                  port_name)

    def unbind_names(self, ip_address, instance_name=None, port_id=None,
                     port_tenant_id=None, device_id=None, device_owner=None,
                     port_name=None):
        if not device_owner:
            return

        ip_alloc = (self.ib_cxt.dhcp_port_ip_alloc
                    if device_owner == n_const.DEVICE_OWNER_DHCP
                    else self.ib_cxt.ip_alloc)
        self._bind_names(ip_alloc.unbind_names, ip_address,
                         instance_name, port_id, port_tenant_id, device_id,
                         device_owner, port_name=port_name, unbind=True)

    def _ensure_dns_zone_availability(self, dns_view, port_tenant_id,
                                      tenant_name, is_external):

        if self.grid_config.dns_support is False:
            return

        forward_zone_eas = eam.get_ea_for_forward_zone(
            self.ib_cxt.user_id, port_tenant_id,
            tenant_name, self.ib_cxt.network,
            self.ib_cxt.subnet,
            self.pattern_builder.get_zone_name(
                is_external=self.ib_cxt.network_is_external))

        dns_zone_name = self.pattern_builder.get_zone_name(
            port_tenant_id=port_tenant_id,
            tenant_name=tenant_name, is_external=is_external)

        ns_group = self.grid_config.ns_group

        if ns_group:
            ib_zone, obj_created = obj.DNSZone.create_check_exists(
                self.ib_cxt.connector,
                view=dns_view,
                ns_group=ns_group,
                fqdn=dns_zone_name,
                extattrs=forward_zone_eas)
            if ib_zone and obj_created:
                LOG.info("Created forward zone: %s with ns_group: %s" % (
                         dns_zone_name, ns_group))

        else:
            self.ib_cxt.reserve_service_members()
            grid_primaries, grid_secondaries = self.ib_cxt.get_dns_members()
            ib_zone, obj_created = obj.DNSZone.create_check_exists(
                self.ib_cxt.connector,
                view=dns_view,
                fqdn=dns_zone_name,
                grid_primary=grid_primaries,
                grid_secondaries=grid_secondaries,
                extattrs=forward_zone_eas)
            if ib_zone and obj_created:
                LOG.info("Created forward zone: %s with "
                         "grid_primaries: %s, grid_secondaries: %s" % (
                             dns_zone_name, grid_primaries, grid_secondaries))

    def _bind_names(self, binding_func, ip_address, instance_name=None,
                    port_id=None, port_tenant_id=None, device_id=None,
                    device_owner=None, ea_ip_address=None, port_name=None,
                    tenant_name=None, unbind=False):
        network_view = self.ib_cxt.mapping.network_view
        dns_view = self.ib_cxt.mapping.dns_view
        is_external = self.ib_cxt.network_is_external

        # It ensure if dns zone is available at NIOS, if not
        # then creates it.
        if not unbind:
            self._ensure_dns_zone_availability(dns_view, port_tenant_id,
                                               tenant_name, is_external)

        fqdn = None
        try:
            fqdn = self.pattern_builder.get_hostname(
                ip_address, instance_name,
                port_id, device_owner,
                device_id, port_name,
                external=is_external,
                port_tenant_id=port_tenant_id,
                tenant_name=tenant_name)
        except ibc_exc.InfobloxConfigException:
            # if unbind - just use fqdn=None to search by ip
            if not unbind:
                raise

        binding_func(network_view, dns_view, ip_address, fqdn, ea_ip_address)



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from infoblox_client import exceptions as ibc_exc
from neutron_lib import constants as n_const

from networking_infoblox.neutron.common import constants as const


class PatternBuilder(object):

    def __init__(self, ib_context):
        self.ib_cxt = ib_context
        self.grid_config = self.ib_cxt.grid_config

    def get_hostname(self, ip_address, instance_name=None, port_id=None,
                     device_owner=None, device_id=None, port_name=None,
                     external=False, port_tenant_id=None, tenant_name=None):
        """Build fqdn based on patterns for network type and device owner.

        Two types of host and domain patterns exist:
        - for external network (optional);
        - for private network (default);
        If pattern for external network is not set, default one is used.
        If device owner is a known one (like dhcp_port, routern interface
        etc.), then per owner patterns are used. Floating ip can be exception
        from this rule if VM is associated with it and instance name is
        present in the pattern.
        """
        if external:
            host_pattern = (self.grid_config.external_host_name_pattern or
                            self.grid_config.default_host_name_pattern)
            domain_pattern = (self.grid_config.external_domain_name_pattern or
                              self.grid_config.default_domain_name_pattern)
        else:
            host_pattern = self.grid_config.default_host_name_pattern
            domain_pattern = self.grid_config.default_domain_name_pattern

        if device_owner in const.NEUTRON_DEVICE_OWNER_TO_PATTERN_MAP and (
                device_owner != n_const.DEVICE_OWNER_FLOATINGIP or not
                instance_name or "{instance_name}" not in host_pattern):
            host_pattern = (
                const.NEUTRON_DEVICE_OWNER_TO_PATTERN_MAP[device_owner])

        pattern = [host_pattern, domain_pattern]
        pattern = '.'.join(el.strip('.') for el in pattern if el)
        return self._build(pattern, ip_address, instance_name, port_id,
                           device_id, port_name=port_name,
                           port_tenant_id=port_tenant_id,
                           tenant_name=tenant_name)

    def get_zone_name_pattern(self, subnet_name=None, is_external=False):
        pattern = self.grid_config.default_domain_name_pattern
        if is_external and self.grid_config.external_domain_name_pattern:
            pattern = self.grid_config.external_domain_name_pattern
        return pattern

    def get_zone_name(self, subnet_name=None, tenant_name=None,
                      port_tenant_id=None, is_external=False):
        pattern = self.get_zone_name_pattern(subnet_name, is_external)
        return self._build(pattern, subnet_name=subnet_name,
                           tenant_name=tenant_name,
                           port_tenant_id=port_tenant_id)

    def _build(self, pattern, ip_address=None, instance_name=None,
               port_id=None, device_id=None, subnet_name=None,
               port_name=None, port_tenant_id=None, tenant_name=None):
        self._validate_pattern(pattern)

        subnet = self.ib_cxt.subnet
        network = self.ib_cxt.network
        if not subnet_name:
            subnet_name = (subnet['name'] if subnet.get('name')
                           else subnet['id'])
        network_name = (network['name'] if network.get('name')
                        else network['id'])

        pattern_dict = {
            'network_id': subnet['network_id'],
            'network_name': network_name,
            'tenant_id': port_tenant_id or self.ib_cxt.tenant_id,
            'tenant_name': tenant_name or self.ib_cxt.tenant_name,
            'subnet_name': subnet_name,
            'subnet_id': subnet['id']
        }

        if port_id:
            pattern_dict['port_id'] = port_id

        if device_id:
            pattern_dict['instance_id'] = device_id
            if instance_name:
                pattern_dict['instance_name'] = re.sub("[^A-Za-z0-9-]", "-",
                                                       instance_name.strip())
            else:
                # During port_creation for instance_name is not available,
                # so set it to instance_id
                pattern_dict['instance_name'] = pattern_dict['instance_id']

        if ip_address:
            octets = ip_address.split('.')
            ip_addr = ip_address.replace('.', '-').replace(':', '-')
            pattern_dict['ip_address'] = ip_addr
            for i in range(len(octets)):
                octet_key = 'ip_address_octet{i}'.format(i=(i + 1))
                pattern_dict[octet_key] = octets[i]

        if port_name:
            pattern_dict['port_name'] = port_name
        elif 'ip_address' in pattern_dict:
            pattern_dict['port_name'] = pattern_dict['ip_address']
        else:
            pattern_dict['port_name'] = port_id

        try:
            # Validate grid config pattern with pattern_dict to
            # restrict user to use only those pattern variable in
            # grid config pattern which is available in pattern_dict
            self._validate_pattern_struct(pattern, pattern_dict)
            fqdn = pattern.format(**pattern_dict)
        except (KeyError, IndexError) as e:
            raise ibc_exc.InfobloxConfigException(
                msg="Invalid pattern %s" % e)
        # Return fqdn as lowercase string as NIOS creates all resources
        # in lowercases.
        return fqdn.lower()

    @staticmethod
    def _validate_pattern_struct(pattern, pattern_dict):
        # This function fetches all variables from grid config
        # pattern and validate the list with supported pattern
        # variables, raises a KeyError if validation fails.
        input_pattern_list = re.findall("\{(.*?)\}", pattern)
        invalid_pattern = set(input_pattern_list) - set(pattern_dict)
        if invalid_pattern:
            raise KeyError(list(invalid_pattern))

    @staticmethod
    def _validate_pattern(pattern):
        invalid_values = ['..']
        for val in invalid_values:
            if val in pattern:
                error_message = "Invalid pattern value {0}".format(val)
                raise ibc_exc.InfobloxConfigException(msg=error_message)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg

from neutron.common import config as common_config

from networking_infoblox.neutron.common import config
from networking_infoblox.neutron.common import grid

from networking_infoblox.tests import base


class GridSyncStub(object):

    def __init__(self, context, connector_fixture):
        self.context = context
        self.fixture = connector_fixture

    def prepare_grid_manager(self, wapi_version):
        self.wapi_version = wapi_version

        self._setup_config()
        self.grid_mgr = grid.GridManager(self.context)
        self.grid_mgr.grid_config.gm_connector = mock.Mock()
        self.grid_mgr.member._discover_dns_settings = mock.Mock(
            return_value=[])
        self.grid_mgr.member._discover_dhcp_settings = mock.Mock(
            return_value=[])

        self._prepare_discovery_resources()

    def get_grid_manager(self):
        return self.grid_mgr

    def _setup_config(self):
        # config init is needed to initialize transport and config loading
        common_config.init([])

        # register infoblox stanza
        config.register_infoblox_ipam_opts(cfg.CONF)
        cfg.CONF.set_override("cloud_data_center_id", 100, 'infoblox')
        cfg.CONF.set_override("ipam_agent_workers", 1, 'infoblox')

        # register infoblox data center stanza
        data_center_id = cfg.CONF.infoblox.cloud_data_center_id
        config.register_infoblox_grid_opts(cfg.CONF, data_center_id)
        data_center = 'infoblox-dc:%s' % data_center_id
        cfg.CONF.set_override('grid_master_host', '192.168.1.7', data_center)
        cfg.CONF.set_override('grid_master_name', 'nios-7.2.0-master.com',
                              data_center)
        cfg.CONF.set_override('data_center_name', 'admin', data_center)
        cfg.CONF.set_override('admin_user_name', 'admin', data_center)
        cfg.CONF.set_override('admin_password', 'infoblox', data_center)
        cfg.CONF.set_override('wapi_version', self.wapi_version, data_center)

    def _prepare_discovery_resources(self):
        resource_map = base.FixtureResourceMap

        member_resource = resource_map.FAKE_MEMBERS_WITHOUT_CLOUD
        if self.grid_mgr.grid_config.is_cloud_wapi:
            member_resource = resource_map.FAKE_MEMBERS_WITH_CLOUD

        license_resource = None
        if self.grid_mgr.grid_config.is_cloud_wapi:
            license_resource = resource_map.FAKE_MEMBER_LICENSES

        netview_resource = resource_map.FAKE_NETWORKVIEW_WITHOUT_CLOUD
        network_resource = resource_map.FAKE_NETWORK_WITHOUT_CLOUD
        if self.grid_mgr.grid_config.is_cloud_wapi:
            netview_resource = resource_map.FAKE_NETWORKVIEW_WITH_CLOUD
            network_resource = resource_map.FAKE_NETWORK_WITH_CLOUD

        dnsview_resource = resource_map.FAKE_DNS_VIEW

        # create members
        member_json = self.fixture.get_object(member_resource)
        self.grid_mgr.member._discover_members = mock.Mock()
        self.grid_mgr.member._discover_members.return_value = member_json

        if license_resource:
            license_json = self.fixture.get_object(license_resource)
            self.grid_mgr.member._discover_member_licenses = mock.Mock()
            self.grid_mgr.member._discover_member_licenses.return_value = (
                license_json)

        config_json = self.fixture.get_object(
            resource_map.FAKE_GRID_MASTER_GRID_CONFIGURATION)
        self.grid_mgr.grid_config._discover_config = mock.Mock()
        self.grid_mgr.grid_config._discover_config.return_value = config_json

        netview_json = self.fixture.get_object(netview_resource)
        self.grid_mgr.mapping._discover_network_views = mock.Mock()
        self.grid_mgr.mapping._discover_network_views.return_value = (
            netview_json)

        network_json = self.fixture.get_object(network_resource)
        self.grid_mgr.mapping._discover_networks = mock.Mock()
        self.grid_mgr.mapping._discover_networks.return_value = network_json

        dnsview_json = self.fixture.get_object(dnsview_resource)
        self.grid_mgr.mapping._discover_dns_views = mock.Mock()
        self.grid_mgr.mapping._discover_dns_views.return_value = dnsview_json



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
eventlet.monkey_patch()

import mock
import time

from neutron.tests.unit import testlib_api
from neutron_lib import context

from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import notification
from networking_infoblox.neutron.common import notification_handler
from networking_infoblox.neutron.common import utils

from networking_infoblox.tests import base
from networking_infoblox.tests.unit import grid_sync_stub


class NotificationTestCase(testlib_api.BaseSqlTestCase, base.RpcTestCase):

    class NotificationEndpointTester(object):
        received_msg_count = 0
        received_payload = []

        def info(self, ctxt, publisher_id, event_type, payload, metadata):
            self.received_payload.append(payload)
            self.received_msg_count += 1

    @mock.patch("neutron.manager.init")
    def setUp(self, initMock):
        super(NotificationTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        stub = grid_sync_stub.GridSyncStub(self.ctx, self.connector_fixture)
        stub.prepare_grid_manager(wapi_version='2.2')
        self.grid_mgr = stub.get_grid_manager()
        self.grid_mgr.grid_config.grid_sync_support = False
        self.grid_mgr.last_sync_time = mock.Mock()
        self.grid_mgr.grid_config.gm_connector = mock.Mock()
        self.grid_mgr.grid_config.gm_connector.wapi_version = '2.0'
        self.event_handler = notification_handler.IpamEventHandler(
            self.ctx, mock.Mock(), self.grid_mgr)

    @mock.patch.object(notification_handler, 'IpamEventHandler')
    def test_notification_endpoint_with_notification_handler(self, mk_ipam_eh):
        msg_context = {}
        publisher_id = 'test_publisher'
        payload = {}
        metadata = {}
        mk_ipam_eh.context = context.get_admin_context()
        endpoint = notification.NotificationEndpoint(self.ctx, None)
        endpoint.handler = self.event_handler

        # go through each event type and verify that each event handler is
        # called from notification handler
        event_types = notification.NotificationEndpoint.event_subscription_list
        for event_type in event_types:
            handler_name = utils.get_notification_handler_name(event_type)
            with mock.patch.object(endpoint.handler,
                                   handler_name) as handler_mock:
                endpoint.info(msg_context, publisher_id, event_type, payload,
                              metadata)
                handler_mock.assert_called_once_with(payload)

    def wait_for_messages(self, endpoint, expected_msg_count):
        while endpoint.received_msg_count < expected_msg_count:
            time.sleep(0.01)

    @mock.patch.object(notification, 'NotificationEndpoint', mock.Mock())
    @mock.patch.object(grid, 'GridManager', mock.Mock())
    def test_notification_service(self):
        publisher_id = 'test_publisher'
        topics = ['notifications']
        notifier = self.get_notifier(topics=topics, publisher_id=publisher_id)

        # prepare service
        service = notification.NotificationService(report_interval=30)
        endpoint = self.NotificationEndpointTester()
        service.event_endpoints = [endpoint]
        service.transport = self.transport
        service.executor = "eventlet"
        service.start()

        # send notification messages
        test_msg_count = 5
        event_type = 'subnet.create.start'
        test_msg_payload = []
        for i in range(test_msg_count):
            test_msg = 'test message %d' % i
            notifier.info({}, event_type, test_msg)
            test_msg_payload.append(test_msg)

        self.wait_for_messages(endpoint, test_msg_count)

        service.stop()

        for i in range(test_msg_count):
            self.assertEqual(test_msg_payload[i], endpoint.received_payload[i])



#!/usr/bin/env python
# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg

from neutron.common import config as common_config
from neutron.conf.agent import common as agent_conf
from neutron_lib import context

from networking_infoblox.neutron.common import config
from networking_infoblox.neutron.common import grid


def register_options():
    agent_conf.register_agent_state_opts_helper(cfg.CONF)
    config.register_infoblox_ipam_opts(cfg.CONF)
    config.register_infoblox_grid_opts(cfg.CONF,
                                       cfg.CONF.infoblox.cloud_data_center_id)


def main():
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    register_options()
    grid.GridManager(context.get_admin_context()).sync(force_sync=True)

if __name__ == "__main__":
    main()



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import constants as n_const
from oslo_log import log as logging
from oslo_utils import uuidutils

from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import utils as ipam_utils

from infoblox_client import exceptions as ib_exc
from infoblox_client import objects as ib_objects

from networking_infoblox._i18n import _LI
from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import context
from networking_infoblox.neutron.common import dns
from networking_infoblox.neutron.common import ea_manager as eam
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import pattern
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi


LOG = logging.getLogger(__name__)


class IpamSyncController(object):

    def __init__(self, ib_context):
        self.ib_cxt = ib_context
        self.grid_config = self.ib_cxt.grid_config
        self.grid_id = self.grid_config.grid_id
        self.pattern_builder = pattern.PatternBuilder(self.ib_cxt)

    def get_subnet(self):
        return self.ib_cxt.ibom.get_network(self.ib_cxt.mapping.network_view,
                                            self.ib_cxt.subnet.get('cidr'))

    def create_subnet(self, rollback_list):
        """Creates subnet equivalent NIOS objects.

        infoblox context contains subnet dictionary from ipam driver and
        network from db query because only subnet data is passed

        The following NIOS objects are created.
        - network view
        - network
        - ip range
        """
        session = self.ib_cxt.context.session
        network = self.ib_cxt.network
        subnet = self.ib_cxt.subnet

        network_id = network.get('id')
        subnet_id = subnet.get('id')

        network_view_exists = (True if self.ib_cxt.mapping.network_view_id
                               else False)

        if self.ib_cxt.mapping.authority_member is None:
            # after authority member reservation, ib_cxt.mapping is updated and
            # its connector and managers are loaded for the new member.
            self.ib_cxt.reserve_authority_member()

        # create a network view if it does not exist
        if not network_view_exists:
            ib_network_view, obj_created = self._create_ib_network_view()
            if ib_network_view and obj_created:
                rollback_list.append(ib_network_view)

        ib_network, created = self._create_ib_network()
        if ib_network:
            if created:
                rollback_list.append(ib_network)
            self._create_ib_ip_range(rollback_list)

        # tenats available only with wapi 2.0+
        features = utils.get_features(self.grid_config.wapi_version)
        if features.tenants:
            ib_tenant = ib_objects.Tenant.search(self.ib_cxt.connector,
                                                 id=self.ib_cxt.tenant_id)
            if (ib_tenant and self.ib_cxt.tenant_name and
                    ib_tenant.name != self.ib_cxt.tenant_name):
                ib_tenant.name = self.ib_cxt.tenant_name
                ib_tenant.update()

        # associate the network view to neutron
        dbi.associate_network_view(session,
                                   self.ib_cxt.mapping.network_view_id,
                                   network_id,
                                   subnet_id)
        return ib_network

    def _create_ib_network_view(self):
        session = self.ib_cxt.context.session

        ea_network_view = eam.get_ea_for_network_view(
            self.ib_cxt.tenant_id,
            self.ib_cxt.tenant_name,
            self.grid_id)
        ib_network_view, obj_created = (
            ib_objects.NetworkView.create_check_exists(
                self.ib_cxt.connector,
                name=self.ib_cxt.mapping.network_view,
                extattrs=ea_network_view))

        network_view_id = utils.get_network_view_id(self.grid_id,
                                                    ib_network_view._ref)
        dbi.update_network_view_id(session,
                                   self.ib_cxt.mapping.network_view_id,
                                   network_view_id)
        self.ib_cxt.mapping.network_view_id = network_view_id

        LOG.info(_LI("Created a network view: %s"), ib_network_view)
        return ib_network_view, obj_created

    def _create_ib_network(self):
        network_view = self.ib_cxt.mapping.network_view
        network = self.ib_cxt.network
        subnet = self.ib_cxt.subnet
        cidr = subnet.get('cidr')
        gateway_ip = subnet.get('gateway_ip')

        ea_network = eam.get_ea_for_network(self.ib_cxt.user_id,
                                            self.ib_cxt.tenant_id,
                                            self.ib_cxt.tenant_name,
                                            network,
                                            subnet)
        network_template = self.grid_config.network_template

        # check if network already exists
        ib_network = ib_objects.Network.search(self.ib_cxt.connector,
                                               network_view=network_view,
                                               cidr=cidr)
        if ib_network:
            if self.ib_cxt.network_is_shared_or_external:
                self.ib_cxt.reserve_service_members(ib_network)
                self.ib_cxt.ibom.update_network_options(ib_network, ea_network)
                LOG.info("ib network already exists so updated options: %s",
                         ib_network)
                return ib_network, False
            raise exc.InfobloxPrivateSubnetAlreadyExist()

        # network creation using template
        if network_template:
            ib_network = self.ib_cxt.ibom.create_network_from_template(
                self.ib_cxt.mapping.network_view, cidr, network_template,
                ea_network)
            self._register_mapping_member()
            self.ib_cxt.reserve_service_members(ib_network)
            self.ib_cxt.ibom.update_network_options(ib_network, ea_network)
            LOG.info("ib network created from template %s: %s",
                     network_template, ib_network)
            return ib_network, True

        # network creation starts
        self.ib_cxt.reserve_service_members()

        relay_trel_ip = None
        ib_network = self.ib_cxt.ibom.create_network(
            self.ib_cxt.mapping.network_view,
            cidr,
            self.ib_cxt.mapping.ib_nameservers,
            self.ib_cxt.mapping.ib_dhcp_members,
            gateway_ip,
            relay_trel_ip,
            ea_network)
        self._register_mapping_member()
        LOG.info("ib network has been created: %s", ib_network)

        self._restart_services()
        return ib_network, True

    def _get_service_members(self, field='member_id'):
        dhcp_members = [m.get(field) for m in self.ib_cxt.mapping.dhcp_members]
        dns_members = [m.get(field) for m in self.ib_cxt.mapping.dns_members]
        service_member_set = set(dhcp_members + dns_members)
        return list(service_member_set)

    def _restart_services(self):
        if self.grid_config.allow_service_restart is False or (
                self.grid_config.dhcp_support is False and
                self.grid_config.dns_support is False):
            return

        member_names = self._get_service_members('member_name')
        for member_name in member_names:
            ib_member = ib_objects.Member.search(self.ib_cxt.connector,
                                                 host_name=member_name,
                                                 return_fields=['host_name'])
            if ib_member:
                self.ib_cxt.ibom.restart_all_services(ib_member)

    def _register_mapping_member(self):
        session = self.ib_cxt.context.session
        mapping_relation = utils.get_mapping_relation(
            self.ib_cxt.mapping.authority_member.member_type)
        mapping_members = dbi.get_mapping_members(
            session,
            self.ib_cxt.mapping.network_view_id,
            self.ib_cxt.mapping.authority_member.member_id,
            self.grid_id,
            mapping_relation)
        if not mapping_members:
            dbi.add_mapping_member(
                session,
                self.ib_cxt.mapping.network_view_id,
                self.ib_cxt.mapping.authority_member.member_id,
                mapping_relation)

    def _create_ib_ip_range(self, rollback_list):
        subnet = self.ib_cxt.subnet
        cidr = subnet.get('cidr')
        ip_version = subnet.get('ip_version')
        gateway_ip = subnet.get('gateway_ip')
        allocation_pools = subnet.get('allocation_pools')
        if not allocation_pools:
            allocation_pools = ipam_utils.generate_pools(cidr, gateway_ip)
        self._allocate_pools(rollback_list, allocation_pools, cidr,
                             ip_version, True)

    def _allocate_pools(self, rollback_list, pools, cidr, ip_version,
                        check_if_exists=False):
        ea_range = eam.get_ea_for_range(self.ib_cxt.user_id,
                                        self.ib_cxt.tenant_id,
                                        self.ib_cxt.tenant_name,
                                        self.ib_cxt.network)
        is_shared = self.ib_cxt.network_is_shared_or_external

        for pool in pools:
            disable = True

            # db_base_plugin uses netaddr but neutronclient uses dict for
            # ip range
            if isinstance(pool, dict) and pool.get('start'):
                start_ip = pool.get('start')
            else:
                start_ip = netaddr.IPAddress(pool.first, ip_version).format()

            if isinstance(pool, dict) and pool.get('end'):
                end_ip = pool.get('end')
            else:
                end_ip = netaddr.IPAddress(pool.last, ip_version).format()

            if check_if_exists:
                ib_ip_range = ib_objects.IPRange.search(
                    self.ib_cxt.connector,
                    network_view=self.ib_cxt.mapping.network_view,
                    network=cidr,
                    start_addr=start_ip,
                    end_addr=end_ip)
                if ib_ip_range:
                    managed = "Unmanaged"  # Not managed by OpenStack
                    if not is_shared or self._range_is_managed(ib_ip_range):
                        managed = "Managed"  # Managed by OpenStack
                        eas = ib_ip_range.extattrs
                        if eas:
                            ea_dict = ib_ip_range.extattrs.ea_dict
                            ea_dict.update(ea_range.ea_dict)
                            ib_ip_range.extattrs = ib_objects.EA(ea_dict)
                        else:
                            ib_ip_range.extattrs = ea_range
                        ib_ip_range.update()
                    LOG.info("%s ip range already existed: %s" %
                             (managed, ib_ip_range))
                    continue

            ib_ip_range = self.ib_cxt.ibom.create_ip_range(
                self.ib_cxt.mapping.network_view,
                start_ip,
                end_ip,
                cidr,
                disable,
                ea_range)
            LOG.info("ip range has been created: %s", ib_ip_range)
            rollback_list.append(ib_ip_range)

    def update_subnet_allocation_pools(self, rollback_list):
        cidr = self.ib_cxt.subnet.get('cidr')
        ip_version = self.ib_cxt.subnet.get('ip_version')
        allocation_pools = self.ib_cxt.subnet.get('allocation_pools')

        # take care of allocation pools
        ib_pools = ib_objects.IPRange.search_all(
            self.ib_cxt.connector,
            network_view=self.ib_cxt.mapping.network_view,
            network=str(cidr))

        added_pool, removed_pool = self._get_changed_pools(
            ib_pools,
            allocation_pools,
            ip_version)

        is_shared = self.ib_cxt.network_is_shared_or_external
        for pool in removed_pool:
            if not is_shared or self._range_is_managed(pool):
                pool.delete()
            else:
                LOG.warning(("Allocation range %s-%s is not managed by " +
                             "OpenStack and is therefore not removed from " +
                             "Infoblox") % (pool.start_addr, pool.end_addr))

        self._allocate_pools(rollback_list, added_pool, cidr, ip_version)
        self._restart_services()

    def update_subnet_details(self, ib_network):
        ea_network = eam.get_ea_for_network(self.ib_cxt.user_id,
                                            self.ib_cxt.tenant_id,
                                            self.ib_cxt.tenant_name,
                                            self.ib_cxt.network,
                                            self.ib_cxt.subnet)

        nameservers = self.ib_cxt.mapping.ib_nameservers
        if nameservers:
            nameservers_option_val = ','.join(nameservers)
            opt_dns = [opt for opt in ib_network.options
                       if opt.name == 'domain-name-servers']
            if opt_dns:
                opt_dns[0].value = nameservers_option_val
            else:
                ib_network.options.append(
                    ib_objects.DhcpOption(name='domain-name-servers',
                                          value=nameservers_option_val))
        else:
            ib_network.options = [opt for opt in ib_network.options
                                  if opt.name != 'domain-name-servers']

        self.ib_cxt.ibom.update_network_options(ib_network, ea_network)
        self._restart_services()

    @staticmethod
    def _get_changed_pools(ib_pools, pools_from_request, ip_version):
        """Calculates difference between nios and neutron ranges.

        :param nios_pools: list of objects.IPRange returned by NIOS
        :param pools_from_request: list of netaddr.IPRange from subnet request
        :return: tuple with two elements:
            - add_list with netaddr.IPRange objects;
            - remove_list with objects.IPRange objects;
        """
        if pools_from_request is None:
            pools_from_request = []

        ib_pool_map = {'%s-%s' % (pool.start_addr, pool.end_addr): pool
                       for pool in ib_pools}

        request_pool_map = {}
        for pool in pools_from_request:
            first_ip = netaddr.IPAddress(pool.first, ip_version).format()
            last_ip = netaddr.IPAddress(pool.last, ip_version).format()
            pool_str = '%s-%s' % (first_ip, last_ip)
            request_pool_map[pool_str] = pool

        old_pools = set(ib_pool_map.keys())
        new_pools = set(request_pool_map.keys())

        added_pools = new_pools - old_pools
        removed_pools = old_pools - new_pools

        added_list = [request_pool_map[pool] for pool in added_pools]
        removed_list = [ib_pool_map[pool] for pool in removed_pools]
        return added_list, removed_list

    def delete_subnet(self, ib_network=None):
        """Frees up resources taken by the subnet and removes the subnet.

        Resources to be released are
        - DHCP/DNS members are freed if not shared by other subnet(s).
        - Network view mapping is removed.
        - Authority member is not freed up for GM but taken care by the agent
          since it is hard to determine at the subnet scope.
        """
        session = self.ib_cxt.context.session
        subnet = self.ib_cxt.subnet
        network = self.ib_cxt.network
        network_view = self.ib_cxt.mapping.network_view

        network_id = network.get('id')
        subnet_id = subnet.get('id')
        cidr = subnet.get('cidr')

        # subnet id is already gone since community plugin already called
        # subnet delete in db level by the time ipam driver reaches here.
        dbi.dissociate_network_view(session, network_id, subnet_id)

        subnet_deletable = (not self.ib_cxt.network_is_shared_or_external or
                            self.grid_config.admin_network_deletion)
        if subnet_deletable:
            ib_ipv4_networks = ib_objects.NetworkV4.search_all(
                self.ib_cxt.connector, network_view=network_view)
            ib_ipv6_networks = ib_objects.NetworkV6.search_all(
                self.ib_cxt.connector, network_view=network_view)
            is_last_subnet_in_netview = (len(ib_ipv4_networks) +
                                         len(ib_ipv6_networks)) == 1

            self._release_service_members(is_last_subnet_in_netview)

            # delete ib network
            self.ib_cxt.ibom.delete_network(network_view, cidr)

            # if no more network exists, remove network view
            if is_last_subnet_in_netview:
                self._remove_network_view()

            self._restart_services()
        else:
            eam.reset_ea_for_network(ib_network)
            ib_network.update()

            ib_ranges = ib_objects.IPRange.search_all(
                self.ib_cxt.connector,
                network_view=network_view,
                network=cidr)
            for ib_range in ib_ranges:
                eam.reset_ea_for_range(ib_range)
                ib_range.update()

    def _release_service_members(self, is_last_subnet_in_netview):
        """Frees up service members

        For CPM, only one subnet should remain in a given network view
        scope since other subnet(s) could use the same dhcp/dns member.

        For GM, dhcp/dns could be GM itself or another member. it is hard
        to determine whether the same member is used in multiple subnets or
        not since we store service members per network view level only.
        for dhcp/dns members, we can release them when only one subnet remains.
        """
        if self.grid_config.dhcp_support is False:
            return

        session = self.ib_cxt.context.session

        if (self.ib_cxt.mapping.authority_member.member_type ==
                const.MEMBER_TYPE_CP_MEMBER):
            if self._is_member_releasable():
                # release service members
                service_member_ids = self._get_service_members('member_id')
                if service_member_ids:
                    dbi.remove_service_members(
                        session,
                        self.ib_cxt.mapping.network_view_id,
                        service_member_ids)
        else:
            # to release dhcp/dns members, only one subnet should remain in the
            # current network view.
            if is_last_subnet_in_netview:
                dhcp_member_ids = [m.member_id
                                   for m in self.ib_cxt.mapping.dhcp_members]
                for member_id in dhcp_member_ids:
                    dbi.remove_service_member(
                        session,
                        self.ib_cxt.mapping.network_view_id,
                        member_id,
                        const.SERVICE_TYPE_DHCP)

                dns_member_ids = [m.member_id
                                  for m in self.ib_cxt.mapping.dns_members]
                for member_id in dns_member_ids:
                    dbi.remove_service_member(
                        session,
                        self.ib_cxt.mapping.network_view_id,
                        member_id,
                        const.SERVICE_TYPE_DNS)

    def _is_member_releasable(self):
        """Determine if service members can be released."""
        session = self.ib_cxt.context.session

        subnet_id = self.ib_cxt.subnet.get('id')
        network_id = self.ib_cxt.subnet.get('network_id')
        tenant_id = self.ib_cxt.subnet.get('tenant_id')

        netview_scope = self.ib_cxt.grid_config.default_network_view_scope

        if netview_scope == const.NETWORK_VIEW_SCOPE_ADDRESS_SCOPE:
            return dbi.is_last_subnet_in_address_scope(session, subnet_id)
        if netview_scope == const.NETWORK_VIEW_SCOPE_TENANT:
            return dbi.is_last_subnet_in_tenant(session, subnet_id, tenant_id)
        if netview_scope == const.NETWORK_VIEW_SCOPE_NETWORK:
            return dbi.is_last_subnet_in_network(session, subnet_id,
                                                 network_id)
        if netview_scope == const.NETWORK_VIEW_SCOPE_SUBNET:
            return True
        return dbi.is_last_subnet(session, subnet_id)

    def _remove_network_view(self):
        session = self.ib_cxt.context.session
        network_view = self.ib_cxt.mapping.network_view

        if (self.grid_config.default_network_view_scope ==
                const.NETWORK_VIEW_SCOPE_SINGLE and
                network_view == self.grid_config.default_network_view):
            return

        # remove ib network view
        self.ib_cxt.ibom.delete_network_view(network_view)

        # release authority member
        dbi.remove_mapping_member(
            session,
            self.ib_cxt.mapping.network_view,
            self.ib_cxt.mapping.authority_member.member_id)

        # remove network view
        dbi.remove_network_views(session,
                                 [self.ib_cxt.mapping.network_view_id])

    def allocate_specific_ip(self, ip_address, mac, port_id=None,
                             port_tenant_id=None, device_id=None,
                             device_owner=None):
        hostname = uuidutils.generate_uuid()

        port_tenant_name = self.ib_cxt.get_tenant_name(port_tenant_id)
        ea_ip_address = eam.get_ea_for_ip(self.ib_cxt.user_id,
                                          port_tenant_id,
                                          port_tenant_name,
                                          self.ib_cxt.network,
                                          port_id,
                                          device_id,
                                          device_owner)
        dns_view = self.ib_cxt.mapping.dns_view
        zone_auth = self.pattern_builder.get_zone_name(
            is_external=self.ib_cxt.network_is_external)

        ip_alloc = (self.ib_cxt.dhcp_port_ip_alloc
                    if device_owner == n_const.DEVICE_OWNER_DHCP
                    else self.ib_cxt.ip_alloc)

        allocated_ip = ip_alloc.allocate_given_ip(
            self.ib_cxt.mapping.network_view,
            dns_view,
            zone_auth,
            hostname,
            mac,
            ip_address,
            ea_ip_address)
        if allocated_ip:
            LOG.info('IP address allocated on Infoblox NIOS: %s',
                     allocated_ip)
            if device_owner == n_const.DEVICE_OWNER_DHCP:
                self.ib_cxt.update_nameservers(allocated_ip)

        return allocated_ip

    def allocate_ip_from_pool(self, subnet_id, allocation_pools, mac,
                              port_id=None, port_tenant_id=None,
                              device_id=None, device_owner=None):
        hostname = uuidutils.generate_uuid()

        port_tenant_name = self.ib_cxt.get_tenant_name(port_tenant_id)
        ea_ip_address = eam.get_ea_for_ip(self.ib_cxt.user_id,
                                          port_tenant_id,
                                          port_tenant_name,
                                          self.ib_cxt.network,
                                          port_id,
                                          device_id,
                                          device_owner)

        dns_view = self.ib_cxt.mapping.dns_view
        zone_auth = self.pattern_builder.get_zone_name(
            is_external=self.ib_cxt.network_is_external)
        allocated_ip = None

        ip_alloc = (self.ib_cxt.dhcp_port_ip_alloc
                    if device_owner == n_const.DEVICE_OWNER_DHCP
                    else self.ib_cxt.ip_alloc)
        for pool in allocation_pools:
            first_ip = pool['start']
            last_ip = pool['end']
            try:
                allocated_ip = ip_alloc.allocate_ip_from_range(
                    self.ib_cxt.mapping.network_view,
                    dns_view,
                    zone_auth,
                    hostname,
                    mac,
                    first_ip,
                    last_ip,
                    ea_ip_address)
                if allocated_ip:
                    break
            except ib_exc.InfobloxCannotAllocateIp:
                LOG.info("Failed to allocate IP from range (%s-%s)." %
                         (first_ip, last_ip))
                continue

            except ib_exc.InfobloxCannotCreateObject:
                LOG.info("Failed to create IP from range (%s-%s)." %
                         (first_ip, last_ip))
                continue

        if allocated_ip:
            LOG.info('IP address allocated on Infoblox NIOS: %s',
                     allocated_ip)
            if device_owner == n_const.DEVICE_OWNER_DHCP:
                self.ib_cxt.update_nameservers(allocated_ip)
        else:
            LOG.info("All IPs from subnet %(subnet_id)s allocated",
                     {'subnet_id': subnet_id})
            raise ipam_exc.IpAddressGenerationFailure(subnet_id=subnet_id)

        return allocated_ip

    def deallocate_ip(self, ip_address):
        dns_view = self.ib_cxt.mapping.dns_view
        self.ib_cxt.ip_alloc.deallocate_ip(self.ib_cxt.mapping.network_view,
                                           dns_view,
                                           ip_address)

    @staticmethod
    def _range_is_managed(ib_range):
        """Determine if an ib_range is managed by OpenStack"""
        return (ib_range and ib_range.extattrs and
                ib_range.extattrs.get(const.EA_CMP_TYPE, "") ==
                const.CLOUD_PLATFORM_NAME)


class IpamAsyncController(object):

    def __init__(self, ib_context):
        self.ib_cxt = ib_context
        self.grid_config = self.ib_cxt.grid_config
        self.grid_id = self.grid_config.grid_id

    def update_network_sync(self, need_new_zones=False):
        """Updates EAs for each subnet that belongs to the updated network."""
        session = self.ib_cxt.context.session
        network = self.ib_cxt.network
        network_id = network.get('id')

        subnets = dbi.get_subnets_by_network_id(session, network_id)
        for subnet in subnets:
            network_view = None
            cidr = subnet.get('cidr')
            subnet_id = subnet.get('id')

            netview_mappings = dbi.get_network_view_mappings(
                session, network_id=network_id, subnet_id=subnet_id)
            if netview_mappings:
                netview_row = utils.find_one_in_list(
                    'id', netview_mappings[0].network_view_id,
                    self.ib_cxt.discovered_network_views)
                network_view = netview_row.network_view

            ib_network = None
            if network_view:
                ib_network = self.ib_cxt.ibom.get_network(network_view, cidr)
                ea_network = eam.get_ea_for_network(self.ib_cxt.user_id,
                                                    self.ib_cxt.tenant_id,
                                                    self.ib_cxt.tenant_name,
                                                    network,
                                                    subnet)
                self.ib_cxt.ibom.update_network_options(ib_network, ea_network)

            if need_new_zones:
                # Need context with ib_network to create zones
                ib_cxt = context.InfobloxContext(
                    self.ib_cxt.context, self.ib_cxt.user_id,
                    network, subnet, self.grid_config,
                    self.ib_cxt.plugin, ib_network=ib_network)
                dns_controller = dns.DnsController(ib_cxt)
                rollback_list = []
                dns_controller.create_dns_zones(rollback_list)
            else:
                self.ib_cxt.subnet = subnet
                dns_controller = dns.DnsController(self.ib_cxt)
                dns_controller.update_dns_zones()

    def update_port_sync(self, port):
        if not port or not port.get('fixed_ips'):
            return

        session = self.ib_cxt.context.session

        for fip in port['fixed_ips']:
            subnet_id = fip['subnet_id']
            ip_address = fip['ip_address']

            netview_mappings = dbi.get_network_view_mappings(
                session,
                network_id=port['network_id'],
                subnet_id=subnet_id)
            if netview_mappings:
                netview_row = utils.find_one_in_list(
                    'id', netview_mappings[0].network_view_id,
                    self.ib_cxt.discovered_network_views)
                network_view = netview_row.network_view

                port_tenant_name = self.ib_cxt.get_tenant_name(
                    port['tenant_id'])
                ea_ip_address = eam.get_ea_for_ip(self.ib_cxt.user_id,
                                                  port['tenant_id'],
                                                  port_tenant_name,
                                                  network_view,
                                                  port['id'],
                                                  port['device_id'],
                                                  port['device_owner'])
                self.ib_cxt.ibom.update_fixed_address_eas(network_view,
                                                          ip_address,
                                                          ea_ip_address)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests.unit import testlib_api
from neutron_lib import context

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import member
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

from networking_infoblox.tests import base
from networking_infoblox.tests.unit import grid_sync_stub


class GridTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(GridTestCase, self).setUp()
        self.ctx = context.get_admin_context()

        self.test_grid_config = grid.GridConfiguration(self.ctx)
        self.test_grid_config.gm_connector = mock.Mock()
        self.test_grid_config.grid_id = 100
        self.test_grid_config.grid_name = "Test Grid 1"
        self.test_grid_config.grid_master_host = '192.168.1.7'
        self.test_grid_config.grid_master_name = 'nios-7.2.0-master.com'
        self.test_grid_config.admin_username = 'admin'
        self.test_grid_config.admin_password = 'infoblox'
        self.test_grid_config.wapi_version = '2.2'

    def _prepare_grid_member(self):
        # create grid
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        # create members
        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        member_mgr._discover_members = mock.Mock(return_value=member_json)
        member_mgr._discover_member_licenses = mock.Mock(return_value=None)
        member_mgr._discover_dns_settings = mock.Mock(return_value=[])
        member_mgr._discover_dhcp_settings = mock.Mock(return_value=[])

        member_mgr.sync_members()

    def test_grid_configuration_without_grid_member(self):
        # grid member sync is required; thus it throws an exception when
        # grid member discovery is not performed
        self.assertRaises(exc.InfobloxCannotFindMember,
                          self.test_grid_config.sync)

    def test_grid_configuration_with_grid_member(self):
        # prepare grid members.
        self._prepare_grid_member()

        # check if GM exists
        db_members = dbi.get_members(self.ctx.session,
                                     grid_id=self.test_grid_config.grid_id,
                                     member_type='GM')
        self.assertEqual('GM', db_members[0]['member_type'])

        # get grid config from GM
        config_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_GRID_MASTER_GRID_CONFIGURATION)
        self.test_grid_config._discover_config = mock.Mock()
        self.test_grid_config._discover_config.return_value = config_json
        self.test_grid_config.sync()

        # verify if grid config object fields are set correctly
        expected = utils.get_ea_value('Default Network View Scope',
                                      config_json)
        self.assertEqual(expected,
                         self.test_grid_config.default_network_view_scope)
        expected = utils.get_ea_value('Default Network View', config_json)
        self.assertEqual(expected,
                         self.test_grid_config.default_network_view)
        expected = utils.get_ea_value('IP Allocation Strategy', config_json)
        self.assertEqual(expected,
                         self.test_grid_config.ip_allocation_strategy)
        expected = utils.get_ea_value('Default Domain Name Pattern',
                                      config_json)
        self.assertEqual(expected,
                         self.test_grid_config.default_domain_name_pattern)
        expected = utils.get_ea_value('Default Host Name Pattern',
                                      config_json)
        self.assertEqual(expected,
                         self.test_grid_config.default_host_name_pattern)
        expected = utils.get_ea_value('DNS Record Binding Types',
                                      config_json)
        self.assertEqual(expected,
                         self.test_grid_config.dns_record_binding_types)

    def test_grid_sync_frequency_check(self):
        # prepare grid manager for sync
        stub = grid_sync_stub.GridSyncStub(self.ctx, self.connector_fixture)
        stub.prepare_grid_manager(wapi_version='1.8')
        grid_mgr = stub.get_grid_manager()
        grid_mgr._report_sync_time = mock.Mock()
        grid_mgr.mapping._sync_nios_for_network_view = mock.Mock()

        # test for no sync
        grid_mgr.grid_config.grid_sync_support = False
        grid_mgr.sync()
        assert not grid_mgr.member._discover_members.called
        assert not grid_mgr.grid_config._discover_config.called
        assert not grid_mgr.mapping._discover_network_views.called
        assert not grid_mgr.mapping._discover_networks.called
        last_sync_time = dbi.get_last_sync_time(self.ctx.session)
        self.assertIsNone(last_sync_time)

        # test for the first sync; expects sync
        grid_mgr.grid_config.grid_sync_support = True
        grid_mgr.sync()
        assert grid_mgr.member._discover_members.called_once
        assert grid_mgr.grid_config._discover_config.called_once
        assert grid_mgr.mapping._discover_network_views.called_once
        assert grid_mgr.mapping._discover_networks.called_once
        sync_time_from_second_sync = dbi.get_last_sync_time(self.ctx.session)
        self.assertEqual(sync_time_from_second_sync, grid_mgr.last_sync_time)

        # test for the second sync; expects no sync due to min wait time
        grid_mgr.grid_config.grid_sync_support = True
        grid_mgr.grid_config.grid_sync_minimum_wait_time = 10
        grid_mgr.sync()
        assert grid_mgr.member._discover_members.called_once
        assert grid_mgr.grid_config._discover_config.called_once
        assert grid_mgr.mapping._discover_network_views.called_once
        assert grid_mgr.mapping._discover_networks.called_once
        # should be the same as last_sync_time from the first sync
        # to prove that dicovery methods are not called in this test case.
        self.assertEqual(sync_time_from_second_sync, grid_mgr.last_sync_time)

        # test for the third sync; expects sync
        grid_mgr.grid_config.grid_sync_support = True
        grid_mgr.grid_config.grid_sync_minimum_wait_time = 0
        grid_mgr.sync()
        assert grid_mgr.member._discover_members.called_once
        assert grid_mgr.grid_config._discover_config.called_once
        assert grid_mgr.mapping._discover_network_views.called_once
        assert grid_mgr.mapping._discover_networks.called_once
        last_sync_time = dbi.get_last_sync_time(self.ctx.session)
        self.assertEqual(last_sync_time, grid_mgr.last_sync_time)

    def _mock_connector(self, get_object=None, create_object=None,
                        delete_object=None):
        connector = mock.Mock()
        connector.get_object.return_value = get_object
        connector.create_object.return_value = create_object
        connector.delete_object.return_value = delete_object
        return connector

    def test_grid_sync_report_sync_time(self):
        # prepare grid manager for sync
        stub = grid_sync_stub.GridSyncStub(self.ctx, self.connector_fixture)
        stub.prepare_grid_manager(wapi_version='1.8')
        grid_mgr = stub.get_grid_manager()
        grid_mgr.mapping._sync_nios_for_network_view = mock.Mock()

        eas = {'Some EA': {'value': 'False'},
               'Zero EA': {'value': '0'}}
        member_mock = {'_ref': ('member/Li5pcHY0X2FkZHJlc3MkMTky'
                                'LjE2OC4xLjEwLzE:192.168.1.10/my_view'),
                       'ipv4_address': '192.168.1.10',
                       'host_name': 'gm',
                       'extattrs': eas}
        connector = self._mock_connector(get_object=[member_mock])
        connector.host = '192.168.1.10'
        grid_mgr.grid_config.gm_connector = connector

        # sync
        grid_mgr.grid_config.grid_sync_support = True
        grid_mgr.grid_config.grid_sync_minimum_wait_time = 0
        grid_mgr.grid_config.report_grid_sync_time = True
        grid_mgr.sync()

        member_ea = member_mock['extattrs'].to_dict()
        sync_info = member_ea[const.EA_LAST_GRID_SYNC_TIME]['value']
        assert isinstance(sync_info, list)
        assert grid_mgr.hostname in sync_info[0]

    def test_grid_sync_report_sync_time_multi_nodes(self):
        # prepare grid manager for sync
        stub = grid_sync_stub.GridSyncStub(self.ctx, self.connector_fixture)
        stub.prepare_grid_manager(wapi_version='1.8')
        grid_mgr = stub.get_grid_manager()
        grid_mgr.mapping._sync_nios_for_network_view = mock.Mock()

        # create a new hostname to report another sync time
        eas = {'Some EA': {'value': 'False'},
               'Zero EA': {'value': '0'},
               'Last Grid Sync Time':
                   {'value': ['controller-1 => 2016-01-28 19:44:56']}}
        member_mock = {'_ref': ('member/Li5pcHY0X2FkZHJlc3MkMTky'
                                'LjE2OC4xLjEwLzE:192.168.1.10/my_view'),
                       'ipv4_address': '192.168.1.10',
                       'host_name': 'gm',
                       'extattrs': eas}
        connector = self._mock_connector(get_object=[member_mock])
        connector.host = '192.168.1.10'
        grid_mgr.grid_config.gm_connector = connector
        grid_mgr.hostname = 'controller-2'

        grid_mgr.grid_config.grid_sync_support = True
        grid_mgr.grid_config.grid_sync_minimum_wait_time = 0
        grid_mgr.grid_config.report_grid_sync_time = True
        grid_mgr.sync()

        member_ea = member_mock['extattrs'].to_dict()
        sync_info = member_ea[const.EA_LAST_GRID_SYNC_TIME]['value']
        assert isinstance(sync_info, list) and len(sync_info) == 2
        assert grid_mgr.hostname in sync_info[1]

    def test__set_default_values(self):
        # validate that default values are set as member properties
        for prop, key in self.test_grid_config.property_to_ea_mapping.items():
            self.assertEqual(const.GRID_CONFIG_DEFAULTS[key],
                             getattr(self.test_grid_config, prop))



# Copyright 2016 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1.identity import generic
from keystoneauth1 import loading

from keystoneclient.v2_0 import client as client_2_0
from keystoneclient.v3 import client as client_3

# from neutron.common import config as cfg
from oslo_config import cfg
from oslo_log import log

from networking_infoblox.neutron.db import infoblox_db as dbi


CONF = cfg.CONF


LOG = log.getLogger(__name__)


def get_identity_service(ib_opts):
    allowed_keystone_version = ['v2.0', 'v3']
    uri_version = ib_opts.keystone_auth_uri.split('/')[-1]
    if uri_version in allowed_keystone_version:
        return ib_opts.keystone_auth_uri, uri_version
    return (ib_opts.keystone_auth_uri + (
        '/%s' % (ib_opts.keystone_auth_version)),
        (ib_opts.keystone_auth_version))


def get_keystone_client():
    key_client = None
    ib_opts = CONF.infoblox
    identity_service, version = get_identity_service(ib_opts)
    auth = generic.Password(
        auth_url=identity_service,
        username=ib_opts.keystone_admin_username,
        password=ib_opts.keystone_admin_password,
        project_name=ib_opts.keystone_admin_project_name,
        user_domain_id=ib_opts.keystone_admin_user_domain_id,
        project_domain_id=ib_opts.keystone_admin_project_domain_id,
        tenant_name=ib_opts.keystone_admin_tenant_name,
        domain_id=ib_opts.keystone_admin_domain_id)

    session = loading.load_session_from_conf_options(CONF, 'infoblox',
                                                     auth=auth)

    if version == 'v2.0':
        key_client = client_2_0.Client(session=session)
    elif version == 'v3':
        key_client = client_3.Client(session=session)
    return key_client


def get_all_tenants():
    try:
        keystone = get_keystone_client()
        if keystone.version == 'v3':
            return keystone.projects.list()
        else:
            return keystone.tenants.list()
    except Exception as e:
        LOG.warning("Could not get tenants due to error: %s", e)
    return []


def update_tenant_mapping(context, networks, tenant_id,
                          tenant_name):
    """Updates tenant_id to tenant_name mapping information.

    Tries to get tenant name to tenant id mapping from context info.
    If context tenant id is different from network tenant_id,
    then check if id to name mapping is already known (check db).
    If id to name mapping still unknown run API call to keystone to
    get all known tenants and store this mapping.
    """

    dbi.add_or_update_tenant(context.session, tenant_id, tenant_name)

    # Get unique tenants ids and check if there are unknown one
    tenant_ids = {net['tenant_id']: True for net in networks}
    if tenant_id in tenant_ids:
        tenant_ids[tenant_id] = False
    unknown_ids = _get_unknown_ids_from_dict(tenant_ids)

    # There are some unknown ids, check if there are mapping in database
    if unknown_ids:
        db_tenants = dbi.get_tenants(context.session,
                                     tenant_ids=unknown_ids)
        for tenant in db_tenants:
            tenant_ids[tenant.tenant_id] = False
        if _get_unknown_ids_from_dict(tenant_ids):
            sync_tenants_from_keystone(context)


def _get_unknown_ids_from_dict(tenant_ids):
    return [id for id, unknown in tenant_ids.items()
            if unknown is True]


def sync_tenants_from_keystone(context):
    tenants = get_all_tenants()
    for tenant in tenants:
        LOG.info("Tenants obtained from keystone: %s", tenant)
        # tenants from keystone have 'id' and 'name' comparing to
        # db cache where 'tenant_id' and 'tenant_name' are used
        dbi.add_or_update_tenant(context.session, tenant.id, tenant.name)
    return len(tenants)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading
from oslo_config import cfg

ipam_opts = [
    cfg.IntOpt('cloud_data_center_id',
               default=0,
               help=_("ID used for selecting a particular grid from one or "
                      "more grids to serve networks in Infoblox backend.")),
    cfg.IntOpt('ipam_agent_workers',
               default=1,
               help=_("Number of Infoblox IPAM agent workers to run")),
    cfg.StrOpt('keystone_auth_uri',
               help=_('Keystone Authtoken URI')),
    cfg.StrOpt('keystone_admin_username',
               help='Admin user name'),
    cfg.StrOpt('keystone_admin_password',
               help='Admin password'),
    cfg.StrOpt('keystone_admin_project_name',
               help='Admin project name'),
    cfg.StrOpt('keystone_admin_tenant_name',
               help='Admin tenant name'),
    cfg.StrOpt('keystone_admin_user_domain_id',
               help='Admin User domain id'),
    cfg.StrOpt('keystone_admin_project_domain_id',
               help='Admin Project domain id'),
    cfg.StrOpt('keystone_admin_domain_id',
               help='Admin Domain id'),
    cfg.StrOpt('keystone_auth_version',
               default='v2.0', help='Auth Version.'),

]


data_center_opts = [
    cfg.StrOpt('data_center_name',
               default='',
               help=_('The name of data center to identify.')),
    cfg.StrOpt('grid_master_host',
               default='',
               help=_('Host IP or name of the grid master.')),
    cfg.StrOpt('grid_master_name',
               default='',
               help=_('Name of the grid master.')),
    cfg.StrOpt('admin_user_name',
               default='',
               help=_("Admin user name to access the grid master or "
                      "cloud platform appliance.")),
    cfg.StrOpt('admin_password',
               default='', secret=True,
               help=_("Admin user password to access the grid master or "
                      "cloud platform appliance.")),
    cfg.StrOpt('wapi_version',
               default='',
               help=_("WAPI (Web API) version.")),
    cfg.StrOpt('ssl_verify',
               default='False',
               help=_("Ensure whether WAPI requests sent over HTTPS require "
                      "SSL verification.")),
    cfg.IntOpt('http_pool_connections',
               default=100,
               help=_("HTTP pool connection.")),
    cfg.IntOpt('http_pool_maxsize',
               default=100,
               help=_("HTTP pool max size.")),
    cfg.IntOpt('http_request_timeout',
               default=120,
               help=_("HTTP request timeout.")),
    cfg.IntOpt('wapi_max_results',
               default=-1000,
               help=_("Maximum number of objects to be returned. If set to a "
                      "negative number the appliance will return an error "
                      "when the number of returned objects would exceed the "
                      "setting. If this is set to a positive number, the "
                      "results will be truncated when necessary."))
]

CONF = cfg.CONF


def register_infoblox_ipam_opts(conf):
    conf.register_group(cfg.OptGroup(
        name='infoblox',
        title="Configuration for Infoblox IPAM Driver"))
    conf.register_opts(ipam_opts, group='infoblox')
    loading.register_auth_conf_options(CONF, 'infoblox')
    loading.register_session_conf_options(CONF, 'infoblox')


def register_infoblox_grid_opts(conf, data_center_id):
    data_center = 'infoblox-dc:%s' % data_center_id
    conf.register_group(cfg.OptGroup(
        name=data_center,
        title="Configuration for Infoblox data center %s" % data_center_id))
    conf.register_opts(data_center_opts, group=data_center)


def get_infoblox_grid_opts(data_center_id):
    grid_info = dict()
    data_center = 'infoblox-dc:%s' % data_center_id
    for opt in data_center_opts:
        grid_info[opt.name] = CONF[data_center][opt.name]
    return grid_info

register_infoblox_ipam_opts(cfg.CONF)
register_infoblox_grid_opts(cfg.CONF, cfg.CONF.infoblox.cloud_data_center_id)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
eventlet.monkey_patch()

import mock

from neutron.tests.unit import testlib_api
from neutron_lib import context
from neutron_lib.plugins import directory

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import context as ib_context
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

from networking_infoblox.tests import base
from networking_infoblox.tests.unit import grid_sync_stub
from networking_infoblox.tests.unit import neutron_plugin_stub


class InfobloxContextTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(InfobloxContextTestCase, self).setUp()
        self.ctx = context.get_admin_context()

        self.setup_coreplugin(neutron_plugin_stub.DB_PLUGIN_KLASS)
        self.plugin = directory.get_plugin()

        self.plugin_stub = neutron_plugin_stub.NeutronPluginStub(self.ctx,
                                                                 self.plugin)

        self.grid_stub = grid_sync_stub.GridSyncStub(self.ctx,
                                                     self.connector_fixture)
        self.grid_stub.prepare_grid_manager(wapi_version='2.2')
        self.grid_mgr = self.grid_stub.get_grid_manager()
        self.grid_mgr._report_sync_time = mock.Mock()
        self.grid_mgr.mapping._sync_nios_for_network_view = mock.Mock()
        self.grid_mgr.sync()

        self.grid_id = self.grid_mgr.grid_config.grid_id
        self.grid_config = self.grid_mgr.grid_config

    def _get_ib_context(self, user_id, network, subnet):
        mock_km_str = 'networking_infoblox.neutron.common.keystone_manager.'
        with mock.patch(mock_km_str + 'get_all_tenants'), (
                mock.patch(mock_km_str + 'sync_tenants_from_keystone')):
            ib_cxt = ib_context.InfobloxContext(self.ctx, user_id,
                                                network, subnet,
                                                self.grid_config, self.plugin)
        return ib_cxt

    def test_network_view_mapping_conditions_with_single_scope(self):
        user_id = 'test user'
        tenant_id = 'test-tenant'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet with cidr tat is not used in mapping conditions
        subnet_name = 'Test Subnet'
        subnet_cidr = '10.0.0.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        # make sure that no subnet cidr is used in mapping conditions
        db_conditions = dbi.get_mapping_conditions(
            self.ctx.session,
            grid_id=self.grid_id,
            neutron_object_value=subnet_cidr)
        self.assertEqual([], db_conditions)

        # check default network view when no mapping condition matches
        self.assertEqual('Single',
                         self.grid_config.default_network_view_scope)
        self.assertEqual('default', self.grid_config.default_network_view)

        # test default mapping as 'Single'
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.connector = mock.Mock()
        ib_cxt.ibom = mock.Mock()
        ib_cxt.ip_allocator = mock.Mock()

        # verify that 'default' view is used
        db_netviews = dbi.get_network_views(self.ctx.session,
                                            grid_id=self.grid_id)
        netview_row = utils.find_one_in_list('network_view', 'default',
                                             db_netviews)
        expected_netview_id = netview_row.id

        db_grid_members = dbi.get_members(self.ctx.session,
                                          grid_id=self.grid_id)
        member_row = utils.find_one_in_list('member_type', 'GM',
                                            db_grid_members)
        expected_member_name = member_row.member_name

        self.assertEqual(expected_netview_id,
                         ib_cxt.mapping.network_view_id)
        self.assertEqual('default',
                         ib_cxt.mapping.network_view)
        self.assertEqual(expected_member_name,
                         ib_cxt.mapping.authority_member.member_name)

    def test_network_view_mapping_conditions_with_tenant_scope(self):
        user_id = 'test user'
        tenant_id = 'test-tenant-id'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet with cidr tat is not used in mapping conditions
        subnet_name = 'Test Subnet'
        subnet_cidr = '10.0.0.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        # make sure that no subnet cidr is used in mapping conditions
        db_conditions = dbi.get_mapping_conditions(
            self.ctx.session,
            grid_id=self.grid_id,
            neutron_object_value=subnet_cidr)
        self.assertEqual([], db_conditions)

        # set default network view scope to 'Tenant'
        self.grid_config.default_network_view_scope = (
            const.NETWORK_VIEW_SCOPE_TENANT)

        # test default mapping as 'Tenant'
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.connector = mock.Mock()
        ib_cxt.ibom = mock.Mock()
        ib_cxt.ip_allocator = mock.Mock()

        # validate the mapping network view
        expected_netview = utils.generate_network_view_name(tenant_id)
        self.assertIsNone(ib_cxt.mapping.network_view_id)
        self.assertEqual(expected_netview, ib_cxt.mapping.network_view)
        self.assertEqual(None, ib_cxt.mapping.authority_member)

    def test_network_view_mapping_conditions_with_tenant_id_condition(self):
        user_id = 'test user'
        tenant_id = '80afaaba012acb9c12888128d5123a09'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet with cidr tat is not used in mapping conditions
        subnet_name = 'Test Subnet'
        subnet_cidr = '10.0.0.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        # make sure that no subnet cidr is used in mapping conditions
        db_conditions = dbi.get_mapping_conditions(
            self.ctx.session,
            grid_id=self.grid_id,
            neutron_object_value=subnet_cidr)
        self.assertEqual([], db_conditions)

        # make sure that tenant id is used in mapping condition once
        db_conditions = dbi.get_mapping_conditions(
            self.ctx.session,
            grid_id=self.grid_id,
            neutron_object_value=tenant_id)
        self.assertEqual(1, len(db_conditions))

        # test mapping where tenant id mapping is found
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.connector = mock.Mock()
        ib_cxt.ibom = mock.Mock()
        ib_cxt.ip_allocator = mock.Mock()

        # validate the mapping network view
        expected_netview_id = db_conditions[0].network_view_id
        db_netviews = dbi.get_network_views(self.ctx.session,
                                            grid_id=self.grid_id)
        netview_row = utils.find_one_in_list('id', expected_netview_id,
                                             db_netviews)
        expected_netview = netview_row.network_view

        db_mapping_members = dbi.get_mapping_members(self.ctx.session,
                                                     expected_netview_id,
                                                     grid_id=self.grid_id)
        expected_member_id = db_mapping_members[0].member_id

        self.assertEqual(expected_netview_id,
                         ib_cxt.mapping.network_view_id)
        self.assertEqual(expected_netview,
                         ib_cxt.mapping.network_view)
        self.assertEqual(expected_member_id,
                         ib_cxt.mapping.authority_member.member_id)

    def test_network_view_mapping_conditions_with_subnet_cidr_condition(self):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet with cidr used in mapping conditions
        subnet_name = 'Test Subnet'
        subnet_cidr = '12.12.2.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        # make sure that mapping condition exists and prepare expectations
        db_conditions = dbi.get_mapping_conditions(
            self.ctx.session,
            grid_id=self.grid_id,
            neutron_object_value=subnet_cidr)
        self.assertEqual(1, len(db_conditions))
        expected_network_view_id = db_conditions[0].network_view_id

        db_network_views = dbi.get_network_views(self.ctx.session,
                                                 grid_id=self.grid_id)
        expected_netview_row = utils.find_one_in_list(
            'id', expected_network_view_id, db_network_views)
        expected_authority_member_id = expected_netview_row.authority_member_id
        expected_network_view = expected_netview_row.network_view

        # prepare network view mapping to neutron network and subnet
        dbi.associate_network_view(
            self.ctx.session, expected_network_view_id, network['id'],
            subnet['id'])

        # test mapping where both tenant id and tenant cidr match
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.connector = mock.Mock()
        ib_cxt.ibom = mock.Mock()
        ib_cxt.ip_allocator = mock.Mock()

        # validate mapping
        self.assertEqual(expected_network_view_id,
                         ib_cxt.mapping.network_view_id)
        self.assertEqual(expected_network_view,
                         ib_cxt.mapping.network_view)
        self.assertEqual(expected_authority_member_id,
                         ib_cxt.mapping.authority_member.member_id)

    @mock.patch.object(dbi, 'add_network_view')
    @mock.patch.object(dbi, 'get_next_authority_member_for_ipam')
    def test_reserve_authority_member_without_dhcp_support(
            self, dbi_next_authority_mock, dbi_network_view_mock):
        user_id = 'test user'
        self.grid_config.dhcp_support = False
        self.grid_config.dns_view = 'test-view'

        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id', 'member_type': 'GM',
             'member_ip': '11.11.1.10', 'member_ipv6': None,
             'member_dns_ip': '12.11.1.11', 'member_dns_ipv6': None,
             'member_dhcp_ip': None, 'member_dhcp_ipv6': None,
             'member_wapi': '11.11.1.10'})
        dbi_next_authority_mock.return_value = test_authority_member

        test_network_view = utils.json_to_obj(
            'NetworkView',
            {'id': 'test-id', 'network_view': 'test-view', 'shared': False})
        dbi_network_view_mock.return_value = test_network_view

        ib_cxt = self._get_ib_context(user_id, None, None)
        ib_cxt.mapping.network_view = test_network_view.network_view

        ib_cxt.reserve_authority_member()

        self.assertEqual(test_network_view.network_view,
                         ib_cxt.mapping.network_view)
        self.assertEqual(test_authority_member.member_id,
                         ib_cxt.mapping.authority_member.member_id)

    @mock.patch.object(dbi, 'add_network_view')
    @mock.patch.object(dbi, 'get_next_authority_member_for_dhcp')
    def test_reserve_authority_member_with_dhcp_support(
            self, dbi_next_authority_mock, dbi_network_view_mock):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.2.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True

        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id', 'member_type': 'CPM',
             'member_ip': '11.11.1.11', 'member_ipv6': None,
             'member_dns_ip': '12.11.1.11', 'member_dns_ipv6': None,
             'member_dhcp_ip': None, 'member_dhcp_ipv6': None,
             'member_name': 'm1', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        dbi_next_authority_mock.return_value = test_authority_member

        test_network_view = utils.json_to_obj(
            'NetworkView',
            {'id': 'ZG5zLm5ldHdvcmtfdmlldyQ1', 'network_view': 'hs-view-1',
             'shared': False})
        dbi_network_view_mock.return_value = test_network_view

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.mapping.network_view = test_network_view.network_view
        ib_cxt.reserve_authority_member()

        self.assertEqual(test_network_view.network_view,
                         ib_cxt.mapping.network_view)
        self.assertEqual(test_authority_member.member_id,
                         ib_cxt.mapping.authority_member.member_id)

    def _test_reserve_service_members_without_ib_network_for_cpm_auth(
            self, test_authority_member, dns_support, expected_dns_members):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.2.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True
        self.grid_config.dns_support = dns_support

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.mapping.authority_member = test_authority_member
        ib_cxt._register_services = mock.Mock()

        ib_cxt.reserve_service_members()

        self.assertEqual([test_authority_member], ib_cxt.mapping.dhcp_members)
        self.assertEqual(expected_dns_members, ib_cxt.mapping.dns_members)

    def test_reserve_service_members_without_ib_network_for_cpm_authortity(
            self):
        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id', 'member_type': 'CPM',
             'member_ip': '11.11.1.11', 'member_ipv6': None,
             'member_dns_ip': '12.11.1.11', 'member_dns_ipv6': None,
             'member_dhcp_ip': None, 'member_dhcp_ipv6': None,
             'member_name': 'm1', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        self._test_reserve_service_members_without_ib_network_for_cpm_auth(
            test_authority_member,
            dns_support=True, expected_dns_members=[test_authority_member])
        self._test_reserve_service_members_without_ib_network_for_cpm_auth(
            test_authority_member,
            dns_support=False, expected_dns_members=[])

    @mock.patch.object(dbi, 'get_service_members')
    def _test_reserve_service_members_without_ib_network_for_gm_auth(
            self, test_authority_member, dns_support, expected_dns_members,
            dbi_service_member_mock):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True
        self.grid_config.dns_support = dns_support

        dbi_service_member_mock.return_value = []

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt.mapping.authority_member = test_authority_member
        ib_cxt.grid_config.use_grid_master_for_dhcp = True
        ib_cxt._register_services = mock.Mock()

        ib_cxt.reserve_service_members()

        self.assertEqual(test_authority_member.member_id,
                         ib_cxt.mapping.dhcp_members[0].member_id)
        self.assertEqual(expected_dns_members, ib_cxt.mapping.dns_members)

    def test_reserve_service_members_without_ib_network_for_gm_authortity(
            self):
        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id-gm', 'member_type': 'GM',
             'member_ip': '11.11.1.11', 'member_ipv6': None,
             'member_dns_ip': '12.11.1.11', 'member_dns_ipv6': None,
             'member_dhcp_ip': None, 'member_dhcp_ipv6': None,
             'member_name': 'gm', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        self._test_reserve_service_members_without_ib_network_for_gm_auth(
            test_authority_member, True, [test_authority_member])
        self._test_reserve_service_members_without_ib_network_for_gm_auth(
            test_authority_member, False, [])

    def _test_reserve_service_members_with_ib_network_gm_owned(
            self, dns_support):
        user_id = 'test user'
        tenant_id = 'tenant-id'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True
        self.grid_config.dns_support = dns_support

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt._register_services = mock.Mock()
        dhcp_members = dbi.get_service_members(
            self.ctx.session,
            network_view_id=ib_cxt.mapping.network_view_id,
            service=const.SERVICE_TYPE_DHCP)
        expected_dhcp_member = dhcp_members[0]

        # ib network with no members and options
        connector = mock.Mock()
        test_ib_network = ib_objects.NetworkV4(connector,
                                               network_view='default',
                                               cidr='12.12.1.0/24')
        test_ib_network.members = []
        test_ib_network.options = []

        ib_cxt.reserve_service_members(test_ib_network)

        expected_dns_members = (
            ib_cxt.mapping.dhcp_members if dns_support else [])
        self.assertEqual(expected_dhcp_member.member_id,
                         ib_cxt.mapping.dhcp_members[0].member_id)
        self.assertEqual(expected_dns_members, ib_cxt.mapping.dns_members)
        actual_opt_router = [opt for opt in test_ib_network.options
                             if opt.name == 'routers']
        self.assertEqual(subnet['gateway_ip'], actual_opt_router[0].value)

    def test_reserve_service_members_with_ib_network_gm_owned(self):
        self._test_reserve_service_members_with_ib_network_gm_owned(True)
        self._test_reserve_service_members_with_ib_network_gm_owned(False)

    def _test_reserve_service_members_with_ib_network_with_dhcp_member(
            self, test_dhcp_member, dns_support, expected_dns_members):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True
        self.grid_config.dns_support = dns_support

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt._register_services = mock.Mock()
        ib_cxt._get_dhcp_members = mock.Mock(return_value=[test_dhcp_member])
        ib_cxt._get_dns_members = mock.Mock(return_value=[test_dhcp_member])

        # ib network with dhcp member assigned
        connector = mock.Mock()
        test_ib_network = ib_objects.NetworkV4(connector,
                                               network_view='test-view',
                                               cidr='12.12.1.0/24')
        test_ib_network.members = [
            ib_objects.AnyMember(_struct='dhcpmember',
                                 name=test_dhcp_member.member_name,
                                 ipv4addr=test_dhcp_member.member_ip)]
        test_ib_network.options = [
            ib_objects.DhcpOption(name='domain-name-servers',
                                  value=test_dhcp_member.member_ip)]

        ib_cxt.reserve_service_members(test_ib_network)

        self.assertEqual([test_dhcp_member], ib_cxt.mapping.dhcp_members)
        self.assertEqual(expected_dns_members, ib_cxt.mapping.dns_members)
        actual_opt_router = [opt for opt in test_ib_network.options
                             if opt.name == 'routers']
        self.assertEqual(subnet['gateway_ip'], actual_opt_router[0].value)

    def test_reserve_service_members_with_ib_network_with_dhcp_member(self):
        test_dhcp_member = utils.json_to_obj(
            'DhcpMember',
            {'member_id': 'member-id', 'member_type': 'REGULAR',
             'member_ip': '11.11.1.12', 'member_ipv6': None,
             'member_dns_ip': '12.11.1.11', 'member_dns_ipv6': None,
             'member_dhcp_ip': None, 'member_dhcp_ipv6': None,
             'member_name': 'm1', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        self._test_reserve_service_members_with_ib_network_with_dhcp_member(
            test_dhcp_member,
            dns_support=True, expected_dns_members=[test_dhcp_member])
        self._test_reserve_service_members_with_ib_network_with_dhcp_member(
            test_dhcp_member,
            dns_support=False, expected_dns_members=[])

    def _test_reserve_service_members_with_ib_network_without_dhcp_member(
            self, dns_support):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True
        self.grid_config.dns_support = dns_support

        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ib_cxt._register_services = mock.Mock()

        # ib network with dhcp member and gateway ips assigned
        connector = mock.Mock()
        test_ib_network = ib_objects.NetworkV4(connector,
                                               network_view='test-view',
                                               cidr='12.12.1.0/24')
        test_ib_network.members = [
            ib_objects.AnyMember(
                _struct='dhcpmember',
                name=ib_cxt.mapping.authority_member.member_name)]
        test_gateway_ip = '12.12.1.1'
        test_ib_network.options = [
            ib_objects.DhcpOption(name='routers', value=test_gateway_ip)]

        ib_cxt.reserve_service_members(test_ib_network)

        expected_dns_members = (
            [ib_cxt.mapping.authority_member] if dns_support else [])
        # authority member is CPM, so dhcp/dns member should be the same as
        # authority member
        self.assertEqual([ib_cxt.mapping.authority_member],
                         ib_cxt.mapping.dhcp_members)
        self.assertEqual(expected_dns_members,
                         ib_cxt.mapping.dns_members)
        actual_opt_router = [opt for opt in test_ib_network.options
                             if opt.name == 'routers']
        self.assertEqual(subnet['gateway_ip'] + ',' + test_gateway_ip,
                         actual_opt_router[0].value)

    def test_reserve_service_members_with_ib_network_without_dhcp_member(self):
        self._test_reserve_service_members_with_ib_network_without_dhcp_member(
            True)
        self._test_reserve_service_members_with_ib_network_without_dhcp_member(
            False)

    def test_get_dns_members_without_dhcp_support(self):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = False

        ib_cxt = self._get_ib_context(user_id, network, subnet)

        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id', 'member_type': 'GM',
             'member_ip': '11.11.1.11', 'member_ipv6': None,
             'member_name': 'm1', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        ib_cxt.mapping.authority_member = test_authority_member

        grid_primaries, grid_secondaries = ib_cxt.get_dns_members()

        expected_grid_primaries = [
            ib_objects.AnyMember(_struct='memberserver',
                                 name=test_authority_member.member_name)]
        self.assertEqual(expected_grid_primaries[0].name,
                         grid_primaries[0].name)
        self.assertEqual(None, grid_secondaries)

    def test_get_dns_members_with_dhcp_support(self):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b7cb98826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dhcp_support = True

        ib_cxt = self._get_ib_context(user_id, network, subnet)

        test_authority_member = utils.json_to_obj(
            'AuthorityMember',
            {'member_id': 'member-id', 'member_type': 'GM',
             'member_ip': '11.11.1.11', 'member_ipv6': None,
             'member_name': 'gm', 'member_status': 'ON',
             'member_wapi': '11.11.1.11'})
        ib_cxt.mapping.authority_member = test_authority_member

        test_dhcp_member_1 = utils.json_to_obj(
            'DhcpMember',
            {'member_id': 'member-id', 'member_type': 'REGULAR',
             'member_ip': '11.11.1.12', 'member_ipv6': None,
             'member_name': 'm1', 'member_status': 'ON',
             'member_wapi': '11.11.1.12'})
        test_dhcp_member_2 = utils.json_to_obj(
            'DhcpMember',
            {'member_id': 'member-id', 'member_type': 'CPM',
             'member_ip': '11.11.1.13', 'member_ipv6': None,
             'member_name': 'm2', 'member_status': 'ON',
             'member_wapi': '11.11.1.13'})
        ib_cxt.mapping.dhcp_members = [test_dhcp_member_1, test_dhcp_member_2]
        ib_cxt.mapping.dns_members = [test_dhcp_member_1, test_dhcp_member_2]

        grid_primaries, grid_secondaries = ib_cxt.get_dns_members()

        expected_primaries = [
            ib_objects.AnyMember(_struct='memberserver',
                                 name=test_authority_member.member_name)]
        expected_secondaries = [
            ib_objects.AnyMember(_struct='memberserver',
                                 name=test_dhcp_member_1.member_name),
            ib_objects.AnyMember(_struct='memberserver',
                                 name=test_dhcp_member_2.member_name)]
        self.assertEqual(expected_primaries, grid_primaries)
        self.assertEqual(expected_secondaries, grid_secondaries)

    def test_get_ip_allocator_for_hosts(self):
        user_id = 'test user'
        tenant_id = '90fbad5a098a4b05524826128d5b40b3'

        # prepare network
        network_name = 'Test Network'
        network = self.plugin_stub.create_network(tenant_id, network_name)

        # prepare subnet
        subnet_name = 'Test Subnet'
        subnet_cidr = '11.11.1.0/24'
        subnet = self.plugin_stub.create_subnet(tenant_id, subnet_name,
                                                network['id'], subnet_cidr)

        self.grid_config.dns_support = True
        self.grid_config.ip_allocation_strategy = (
            const.IP_ALLOCATION_STRATEGY_HOST_RECORD)
        self.grid_config.zone_creation_strategy = (
            const.GRID_CONFIG_DEFAULTS[
                const.EA_GRID_CONFIG_ZONE_CREATION_STRATEGY])
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ip_allocator = ib_cxt._get_ip_allocator()
        self.assertEqual(True, ip_allocator.opts['configure_for_dns'])

        self.grid_config.dns_support = False
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ip_allocator = ib_cxt._get_ip_allocator()
        self.assertEqual(False, ip_allocator.opts['configure_for_dns'])

        self.grid_config.dns_support = True
        self.grid_config.zone_creation_strategy = [
            const.ZONE_CREATION_STRATEGY_REVERSE]
        ib_cxt = self._get_ib_context(user_id, network, subnet)
        ip_allocator = ib_cxt._get_ip_allocator()
        self.assertEqual(False, ip_allocator.opts['configure_for_dns'])



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import constants
from networking_infoblox.neutron.common import context
from networking_infoblox.neutron.common import dns
from networking_infoblox.neutron.common import ipam
from networking_infoblox.neutron.common import notification_handler as handler
from networking_infoblox.neutron.common import pattern
from networking_infoblox.neutron.db import infoblox_db as dbi
from networking_infoblox.tests import base


class TestIpamEventHandler(base.TestCase):

    def setUp(self):
        super(TestIpamEventHandler, self).setUp()
        self.prepare_ipam_event_handler()

    @mock.patch("neutron.manager.init")
    def prepare_ipam_event_handler(self, initMock):
        self.context = mock.Mock()
        self.plugin = mock.Mock()
        self.grid_manager = mock.Mock()
        self.grid_manager.grid_config.gm_connector = mock.Mock()
        self.grid_manager.grid_config.gm_connector.wapi_version = '2.0'
        self.grid_manager.grid_config.zone_creation_strategy = (
            constants.GRID_CONFIG_DEFAULTS[
                constants.EA_GRID_CONFIG_ZONE_CREATION_STRATEGY])

        self.ipam_handler = handler.IpamEventHandler(self.context, self.plugin,
                                                     self.grid_manager)
        self.ipam_handler.user_id = 'test-user'
        self.ipam_handler._resync = mock.Mock()

    def test_create_network_alert_should_call_resync(self):
        payload = {'network': {}}
        self.ipam_handler.create_network_alert(payload)
        self.ipam_handler._resync.assert_called_once_with()

    def test_create_subnet_alert_should_call_resync(self):
        payload = {'subnet': {}}
        self.ipam_handler.create_subnet_alert(payload)
        self.ipam_handler._resync.assert_called_once_with()

    def test_update_network_sync(self):
        payload = {'network': {}}
        with mock.patch.object(ipam.IpamAsyncController,
                               'update_network_sync') as controller_mock:
            self.ipam_handler.update_network_sync(payload)
        controller_mock.assert_called_once_with(False)

    @mock.patch.object(dbi, 'get_network', mock.Mock())
    @mock.patch.object(dbi, 'add_or_update_network', mock.Mock())
    def test_update_network_sync_name_not_changed(self):
        net_id = 'net_id'
        net_name = 'new_name'
        payload = {'network': {'name': net_name, 'id': net_id}}
        old_network = mock.Mock()
        old_network.network_name = net_name
        dbi.get_network.return_value = old_network
        with mock.patch.object(ipam.IpamAsyncController,
                               'update_network_sync') as controller_mock:
            self.ipam_handler.update_network_sync(payload)
        controller_mock.assert_called_once_with(False)
        dbi.get_network.assert_called_once_with(mock.ANY, net_id)
        dbi.add_or_update_network.assert_not_called()

    @mock.patch.object(dbi, 'get_network')
    @mock.patch.object(dbi, 'add_or_update_network')
    def _test_update_network_name(self, pattern, need_zones,
                                  add_or_update_network_mock,
                                  get_network_mock):
        net_id = 'net_id'
        net_name = 'new_name'
        payload = {'network': {'name': net_name, 'id': net_id}}
        old_network = mock.Mock()
        old_network.network_name = 'old_name'
        get_network_mock.return_value = old_network
        grid_config = self.grid_manager.grid_config
        grid_config.default_domain_name_pattern = pattern
        with mock.patch.object(ipam.IpamAsyncController,
                               'update_network_sync') as controller_mock:
            self.ipam_handler.update_network_sync(payload)
        controller_mock.assert_called_once_with(need_zones)
        get_network_mock.assert_called_once_with(mock.ANY, net_id)
        add_or_update_network_mock.assert_called_once_with(
            mock.ANY, net_id, net_name)

    def test_network_name_changed_zones_needed(self):
        self._test_update_network_name('{network_name}', True)

    def test_network_name_changed_zones_not_needed(self):
        self._test_update_network_name('{subnet_name}', False)

    def test_create_subnet_sync_should_call_resync(self):
        payload = {'subnet': {}}
        self.ipam_handler.create_subnet_sync(payload)
        self.ipam_handler._resync.assert_called_once_with(True)

    def test_delete_subnet_sync(self):
        payload = {'subnet_id': 'subnet-id'}
        self.ipam_handler.delete_subnet_sync(payload)
        self.ipam_handler._resync.assert_called_once_with(True)

    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    @mock.patch.object(pattern.PatternBuilder,
                       "_validate_pattern_struct", mock.Mock())
    def test_update_floatingip_sync(self, ib_cxt_mock):
        payload = {'floatingip': {'id': 'floatingip-id',
                                  'tenant_id': 'tenant-id',
                                  'port_id': 'port-id',
                                  'floating_network_id': 'floating-network-id',
                                  'floating_ip_address': '8.8.8.3'}}
        instance_name = 'test-inst'
        self.ipam_handler._get_instance_name_from_fip = (
            mock.Mock(return_value=instance_name))
        ib_cxt_mock.grid_config.default_domain_name_pattern = 'global.com'
        self.ipam_handler._get_mapping_neutron_subnet = mock.Mock()
        with mock.patch.object(dns.DnsController,
                               'bind_names') as bind_name_mock:
            self.ipam_handler.update_floatingip_sync(payload)
        bind_name_mock.assert_called_with(
            payload['floatingip']['floating_ip_address'],
            instance_name,
            mock.ANY,
            payload['floatingip']['tenant_id'],
            mock.ANY,
            mock.ANY,
            True,
            mock.ANY)

    @mock.patch.object(dbi, 'get_instance', mock.Mock())
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_get_instance_name_from_fip(self, ib_cxt_mock):
        floatingip = {'id': 'floatingip-id',
                      'tenant_id': 'tenant-id',
                      'port_id': 'port-id',
                      'floating_network_id': 'floating-network-id',
                      'floating_ip_address': '8.8.8.3',
                      'fixed_ip_address': '1.1.1.1'}
        port = {'fixed_ips': [{'subnet_id': 'subnet-id',
                               'ip_address': floatingip['fixed_ip_address']}],
                'device_id': 'instance-id1',
                'device_owner': 'compute:nova'}
        instance_name = 'test-inst'
        extattrs = mock.Mock(extattrs={'VM Name': instance_name})
        dbi.get_instance.return_value = None
        self.plugin.get_port = mock.Mock(return_value=port)
        self.plugin.get_subnet = mock.Mock()
        with mock.patch.object(ib_objects.FixedAddress,
                               'search',
                               return_value=extattrs):
            self.assertEqual(
                instance_name,
                self.ipam_handler._get_instance_name_from_fip(floatingip))
            self.plugin.get_port.assert_called_with(mock.ANY,
                                                    floatingip['port_id'])
            self.plugin.get_subnet.assert_called_with(
                mock.ANY, port['fixed_ips'][0]['subnet_id'])
            # now check with instance in db
            dbi.get_instance.return_value = mock.Mock()
            dbi.get_instance.return_value.instance_name = instance_name
            self.plugin.get_port.reset_mock()
            self.plugin.get_subnet.reset_mock()
            self.assertEqual(
                instance_name,
                self.ipam_handler._get_instance_name_from_fip(floatingip))
            self.plugin.get_port.assert_called_with(mock.ANY,
                                                    floatingip['port_id'])
            self.plugin.get_subnet.assert_not_called()

    def _prepare_context(self):
        message_context = {'project_name': u'admin',
                           'user_id': u'9510723b6555473cb735ce6640e680cb',
                           'show_deleted': False,
                           'roles': [u'admin'],
                           'tenant_name': u'admin',
                           'auth_token': u'dd7257d2e02f41c2908ffdc197af9061',
                           'tenant_id': u'bf0806763e32436bbdb8fd9b6ebfac93',
                           'tenant': u'bf0806763e32436bbdb8fd9b6ebfac93',
                           'user_name': u'admin'}
        self.ipam_handler.ctxt = message_context

    def test_create_network_sync_same_tenant(self):
        payload = {
            'network': {'status': 'ACTIVE',
                        'subnets': [],
                        'name': 'network_name',
                        'tenant_id': 'bf0806763e32436bbdb8fd9b6ebfac93'}}
        self._prepare_context()
        self.ipam_handler.create_network_sync(payload)

    @mock.patch('networking_infoblox.neutron.db.infoblox_db.get_tenants')
    def test_create_network_sync_tenant_mismatch(self, get_mock):
        db_tenant = mock.Mock()
        db_tenant.id = '25ba7c0a7b1e4266b48a7731b2502e05'
        get_mock.return_value = [mock.Mock()]
        payload = {
            'network': {'status': 'ACTIVE',
                        'subnets': [],
                        'name': 'network_name',
                        'tenant_id': db_tenant.id}}
        self._prepare_context()
        with mock.patch('networking_infoblox.neutron.common.'
                        'keystone_manager.update_tenant_mapping'):
            self.ipam_handler.create_network_sync(payload)

    @mock.patch.object(dbi, 'add_or_update_instance', mock.Mock())
    def test_create_instance_sync_instance_name_create(self):
        instance_id = 'instance-id'
        instance_name = 'test-host'
        payload = {
            'instance_id': instance_id,
            'hostname': instance_name,
            'fixed_ips': {}
            }
        self._prepare_context()
        self.ipam_handler.create_instance_sync(payload)
        dbi.add_or_update_instance.assert_called_once_with(
            self.context.session, instance_id, instance_name)

    @mock.patch.object(dbi, 'remove_instance', mock.Mock())
    @mock.patch.object(dbi, 'get_external_subnets', mock.Mock())
    def test_delete_instance_sync_instance_name_delete(self):
        instance_id = 'instance-id'
        payload = {
            'instance_id': instance_id,
            }
        self._prepare_context()
        dbi.get_external_subnets.return_value = ()
        self.ipam_handler.delete_instance_sync(payload)
        dbi.remove_instance.assert_called_once_with(
            self.context.session, instance_id)

    @mock.patch.object(dbi, 'add_or_update_instance', mock.Mock())
    @mock.patch.object(context, 'InfobloxContext', mock.Mock())
    @mock.patch.object(dns, 'DnsController', mock.Mock())
    def test_create_instance_bind_names(self):
        instance_id = 'instance-id'
        instance_name = 'test-host'
        dns_controller = mock.MagicMock()
        dns.DnsController.return_value = dns_controller
        payload = {
            'instance_id': instance_id,
            'hostname': instance_name,
            'fixed_ips': [
                {'address': 'adr-1', 'subnet_id': 'subnet-1'},
                {'address': 'adr-2', 'subnet_id': 'subnet-2'},
                {'address': 'adr-3', 'subnet_id': 'subnet-3'},
                {'address': 'adr-4', 'subnet_id': 'subnet-4'},
            ]
        }
        self._prepare_context()
        ports = [
            {
                'id': 'port-id-1',
                'name': 'port-name-1',
                'tenant_id': 'tenant-id-1',
                'device_id': 'device-id-1',
                'device_owner': 'device-owner-1',
                'fixed_ips': [
                    {'ip_address': 'adr-1', 'subnet_id': 'subnet-1'},
                    {'ip_address': 'adr-2', 'subnet_id': 'subnet-2'},
                ]
            },
            {
                'id': 'port-id-2',
                'name': 'port-name-2',
                'tenant_id': 'tenant-id-2',
                'device_id': 'device-id-2',
                'device_owner': 'device-owner-2',
                'fixed_ips': [
                    {'ip_address': 'adr-3', 'subnet_id': 'subnet-3'},
                    {'ip_address': 'adr-4', 'subnet_id': 'subnet-4'},
                ]
            },
        ]
        with mock.patch.object(self.ipam_handler, 'plugin'):
            self.ipam_handler.plugin.get_ports.return_value = ports
            self.ipam_handler.create_instance_sync(payload)
            assert dns_controller.method_calls == [
                mock.call.bind_names(
                    'adr-1', 'test-host', 'port-id-1', 'tenant-id-1',
                    'device-id-1', 'device-owner-1', port_name='port-name-1'),
                mock.call.bind_names(
                    'adr-2', 'test-host', 'port-id-1', 'tenant-id-1',
                    'device-id-1', 'device-owner-1', port_name='port-name-1'),
                mock.call.bind_names(
                    'adr-3', 'test-host', 'port-id-2', 'tenant-id-2',
                    'device-id-2', 'device-owner-2', port_name='port-name-2'),
                mock.call.bind_names(
                    'adr-4', 'test-host', 'port-id-2', 'tenant-id-2',
                    'device-id-2', 'device-owner-2', port_name='port-name-2')
            ]

    @mock.patch.object(dbi, 'get_instance', mock.Mock())
    @mock.patch.object(context, 'InfobloxContext', mock.Mock())
    @mock.patch.object(dns, 'DnsController', mock.Mock())
    def test_update_port_sync(self):
        subnet_id = 'test-subnet'
        port_name = 'test-port'
        port_id = 'test-port-id'
        ip_address = '192.168.126.4'
        instance_id = 'test-instance-id'
        instance_name = 'test-instance'
        network_id = 'test-network-id'
        tenant_id = 'test-tenant-id'
        nova = constants.NEUTRON_DEVICE_OWNER_COMPUTE_NOVA
        payload_attach = {
            'port': {
                'device_owner': nova,
                'fixed_ips': [{'subnet_id': subnet_id,
                               'ip_address': ip_address}],
                'id': port_id,
                'binding:vif_type': 'ovs',
                'device_id': instance_id,
                'name': port_name,
                'network_id': network_id,
                'tenant_id': tenant_id,
            }
        }
        payload_detach = {
            'port': {
                'device_owner': '',
                'fixed_ips': [{'subnet_id': subnet_id,
                               'ip_address': ip_address}],
                'id': port_id,
                'binding:vif_type': 'unbound',
                'device_id': '',
                'name': port_name,
                'network_id': network_id,
                'tenant_id': tenant_id,
            }
        }
        dns_controller = mock.MagicMock()
        dns.DnsController.return_value = dns_controller
        instance = mock.MagicMock()
        instance.instance_name = instance_name
        dbi.get_instance.return_value = instance
        # Attach port and ensure that bind_names called
        self.ipam_handler.update_port_sync(payload_attach)
        dns_controller.method_calls == [
            mock.call.bind_names(
                ip_address, instance_name, port_id, tenant_id, instance_id,
                nova, port_name=port_name)]
        # now detach port and ensure that unbind_names called
        self.ipam_handler.update_port_sync(payload_detach)
        dns_controller.method_calls == [
            mock.call.bind_names(
                ip_address, instance_name, port_id, tenant_id, instance_id,
                nova, port_name=port_name),
            mock.call.unbind_names(
                ip_address, None, port_id, tenant_id, None, 'compute:nova',
                port_name=port_name)
        ]



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import ip_allocator
from networking_infoblox.tests import base


class FixedAddressAllocatorTestCase(base.TestCase):
    def setUp(self):
        super(FixedAddressAllocatorTestCase, self).setUp()
        self.ib_mock = mock.Mock()

        self.extattrs = 'test-extattrs'
        self.netview = 'some-test-net-view'
        self.mac = 'de:ad:be:ef:00:00'
        self.ip = '192.168.1.1'
        self.dnsview = 'some-dns-view'
        self.zone_auth = 'zone-auth'
        self.hostname = 'host1'
        self.dhcp_enabled = True

        options = {'use_host_record': False}

        self.allocator = ip_allocator.IPAllocator(self.ib_mock, options)

    def test_creates_fixed_address_on_allocate_ip(self):
        self.ib_mock.get_fixed_addresses_by_mac.return_value = []
        self.allocator.allocate_given_ip(
            self.netview, self.dnsview, self.zone_auth,
            self.hostname, self.mac, self.ip, self.extattrs)

        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_for_given_ip.assert_called_once_with(
            self.netview, self.mac, self.ip, self.extattrs)

    def test_creates_fixed_address_on_allocate_ip_reuse_failed(self):
        fa = mock.Mock()
        fa.ip = self.ip + '1'
        self.ib_mock.get_fixed_addresses_by_mac.return_value = [fa]
        self.allocator.allocate_given_ip(
            self.netview, self.dnsview, self.zone_auth,
            self.hostname, self.mac, self.ip, self.extattrs)

        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_for_given_ip.assert_called_once_with(
            self.netview, self.mac, self.ip, self.extattrs)

    def test_creates_fixed_address_on_allocate_ip_reuse_success(self):
        fa = mock.Mock()
        fa.ip = self.ip
        self.ib_mock.get_fixed_addresses_by_mac.return_value = [fa]
        self.allocator.allocate_given_ip(
            self.netview, self.dnsview, self.zone_auth,
            self.hostname, self.mac, self.ip, self.extattrs)

        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_for_given_ip.assert_not_called()

    def test_creates_fixed_address_range_on_range_allocation(self):
        first_ip = '192.168.1.1'
        last_ip = '192.168.1.123'
        self.ib_mock.get_fixed_addresses_by_mac.return_value = []

        self.allocator.allocate_ip_from_range(
            self.netview, self.dnsview, self.zone_auth, self.hostname,
            self.mac, first_ip, last_ip, self.extattrs)

        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_from_range.assert_called_once_with(
            self.netview, self.mac, first_ip, last_ip, self.extattrs)

    def test_allocate_ip_from_range_reuse_failed(self):
        first_ip = '192.168.1.1'
        last_ip = '192.168.1.123'
        fa = mock.Mock()
        fa.ip = '192.168.1.124'
        self.ib_mock.get_fixed_addresses_by_mac.return_value = [fa]

        ip = self.allocator.allocate_ip_from_range(
            self.netview, self.dnsview, self.zone_auth, self.hostname,
            self.mac, first_ip, last_ip, self.extattrs)

        assert ip != fa.ip
        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_from_range.assert_called_once_with(
            self.netview, self.mac, first_ip, last_ip, self.extattrs)

    def test_allocate_ip_from_range_reuse_success(self):
        first_ip = '192.168.1.1'
        last_ip = '192.168.1.123'
        fa = mock.Mock()
        fa.ip = '192.168.1.12'
        self.ib_mock.get_fixed_addresses_by_mac.return_value = [fa]

        ip = self.allocator.allocate_ip_from_range(
            self.netview, self.dnsview, self.zone_auth, self.hostname,
            self.mac, first_ip, last_ip, self.extattrs)

        assert ip == fa.ip
        self.ib_mock.get_fixed_addresses_by_mac.assert_called_once_with(
            self.netview, self.mac)
        self.ib_mock.create_fixed_address_from_range.assert_not_called()

    def test_deletes_fixed_address(self):
        self.allocator.deallocate_ip(self.netview, self.dnsview, self.ip)

        self.ib_mock.delete_fixed_address.assert_called_once_with(self.netview,
                                                                  self.ip)


class HostRecordAllocatorTestCase(base.TestCase):

    def _test_creates_host_record_on_allocate_ip(self, use_dhcp, use_dns):
        ib_mock = mock.MagicMock()

        netview = 'some-test-net-view'
        dnsview = 'some-dns-view'
        zone_auth = 'zone-auth'
        hostname = 'host1'
        mac = 'de:ad:be:ef:00:00'
        ip = '192.168.1.1'

        ib_mock.find_hostname.return_value = None
        ib_mock.find_host_records_by_mac.return_value = None
        options = {'use_host_record': True,
                   'configure_for_dhcp': use_dhcp,
                   'configure_for_dns': use_dns}

        allocator = ip_allocator.IPAllocator(ib_mock, options)
        allocator.allocate_given_ip(netview, dnsview, zone_auth, hostname,
                                    mac, ip)

        ib_mock.create_host_record_for_given_ip.assert_called_once_with(
            dnsview, zone_auth, hostname, mac, ip, mock.ANY, use_dhcp, use_dns)

    def test_creates_host_record_on_allocate_ip_use_dhcp(self):
        self._test_creates_host_record_on_allocate_ip(True, True)

    def test_creates_host_record_on_allocate_ip_no_dhcp(self):
        self._test_creates_host_record_on_allocate_ip(False, False)

    def _test_creates_host_record_range_on_range_allocation(self, use_dhcp,
                                                            use_dns):
        ib_mock = mock.MagicMock()

        netview = 'some-test-net-view'
        dnsview = 'some-dns-view'
        zone_auth = 'zone-auth'
        hostname = 'host1'
        mac = 'de:ad:be:ef:00:00'
        first_ip = '192.168.1.2'
        last_ip = '192.168.1.254'

        ib_mock.find_hostname.return_value = None
        ib_mock.find_host_records_by_mac.return_value = None
        options = {'use_host_record': True,
                   'configure_for_dhcp': use_dhcp,
                   'configure_for_dns': use_dns}

        allocator = ip_allocator.IPAllocator(ib_mock, options)
        allocator.allocate_ip_from_range(
            netview, dnsview, zone_auth, hostname, mac, first_ip, last_ip)

        ib_mock.create_host_record_from_range.assert_called_once_with(
            dnsview, netview, zone_auth, hostname,
            mac, first_ip, last_ip, mock.ANY, use_dhcp, use_dns)

    def test_creates_host_record_range_on_range_allocation_use_dhcp(self):
        self._test_creates_host_record_range_on_range_allocation(True, True)

    def _test_creates_host_record_range_on_range_allocation_no_dhcp(self):
        self._test_creates_host_record_range_on_range_allocation(False, False)

    def test_deletes_host_record(self):
        ib_mock = mock.MagicMock()

        netview = 'some-test-net-view'
        dnsview = 'some-dns-view'
        ip_1 = ['192.168.1.2', 'de:ad:be:ef:00:00']
        ip_2 = ['192.168.1.3', 'ff:ee:be:ae:12:00']

        options = {'use_host_record': True}

        allocator = ip_allocator.IPAllocator(ib_mock, options)
        host_record_mock = mock.Mock()
        ip_obj_1 = ib_objects.IP.create(ip=ip_1[0], mac=ip_1[1])
        ip_obj_2 = ib_objects.IP.create(ip=ip_2[0], mac=ip_2[1])
        host_record_mock.ip = [ip_obj_1, ip_obj_2]
        allocator.manager.get_host_record = mock.Mock()
        allocator.manager.get_host_record.return_value = host_record_mock

        allocator.deallocate_ip(netview, dnsview, ip_1[0])

        ib_mock.delete_ip_from_host_record.assert_called_once_with(
            host_record_mock, ip_1[0])

    def _check_bind_names_calls(self, opts, params, expected_find,
                                expected_get_host, expected_bind):
        ib_mock = mock.MagicMock()
        ib_mock.find_hostname.return_value = None
        ib_mock.get_host_record.return_value = None
        allocator = ip_allocator.IPAllocator(ib_mock, opts)
        allocator.bind_names(*params)
        ib_mock.find_hostname.assert_called_once_with(*expected_find)
        ib_mock.get_host_record.assert_called_once_with(*expected_get_host)
        ib_mock.bind_name_with_host_record.assert_called_once_with(
            *expected_bind)

    def test_bind_names_for_non_dns(self):
        netview = 'some-test-net-view'
        dns_view = 'some-dns-view'
        hostname = 'host1'
        ip = '192.168.1.2'
        extattrs = 'test-extattrs'

        options = {'use_host_record': True,
                   'configure_for_dhcp': True,
                   'configure_for_dns': True}

        # First bind dns host - should be used provided dns view name
        self._check_bind_names_calls(
            options,
            [netview, dns_view, ip, hostname, extattrs],
            [dns_view, hostname, ip, None],  # expected find_hostname params
            [dns_view, ip, None],  # expected get_host_record params
            [dns_view, ip, hostname, extattrs, None])  # expected for bind_name

        options['configure_for_dns'] = False
        # Now bind non-dns host in all network view - network view
        # name should be used, dns view name should be not used
        self._check_bind_names_calls(
            options,
            [netview, dns_view, ip, hostname, extattrs],
            [None, hostname, ip, netview],  # expected find_hostname params
            [None, ip, netview],  # expected get_host_record params
            [None, ip, hostname, extattrs, netview])  # expected for bind_name
